from moulti.helpers import abridge_dict, abridge_string

def test_abridge_string():
	assert abridge_string('too small, not abridged') == 'too small, not abridged'
	assert abridge_string('ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789', 40) == 'ABCDEFGHIJKL...Zabcdefghij...z0123456789'
	assert abridge_string('ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789', 40, '.....') == 'ABCDEFGHIJ.....abcdefghij.....0123456789'

def test_abridge_dict():
	original_dict = {
		'foo': 'too small, not abridged',
		'bar': 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789',
		'int': 1,
		'float': 1.1,
		'bool': True,
		'array': [
			'recursive? no, this function is not expected to work recursively',
		],
		'dict': {},
	}
	expected_dict = {
		'foo': 'too small, not abridged',
		'bar': 'ABCDEFGHIJKL...Zabcdefghij...z0123456789',
		'int': 1,
		'float': 1.1,
		'bool': True,
		'array': [
			'recursive? no, this function is not expected to work recursively',
		],
		'dict': {},
	}
	assert abridge_dict(original_dict, 40) == expected_dict



from .common import moulti_test
assert moulti_test

def do_test(moulti_test, position, direction):
	command = ['display-steps.bash', '--step-position', position, '--step-direction', direction]
	assert moulti_test(command=command)

def test_display_steps_top_down(moulti_test):
	do_test(moulti_test, 'top', 'down')

def test_display_steps_top_up(moulti_test):
	do_test(moulti_test, 'top', 'up')

def test_display_steps_bottom_down(moulti_test):
	do_test(moulti_test, 'bottom', 'down')

def test_display_steps_bottom_up(moulti_test):
	do_test(moulti_test, 'bottom', 'up')



from typing import Any
from typing_extensions import Self
from textual.app import ComposeResult
from textual.events import Click
from textual.widget import Widget
from textual.widgets import Static, Collapsible
from moulti.search import TextSearch
from ..abstractstep.tui import AbstractStep

SEARCH_RESET = 0
SEARCH_TITLE = 1
SEARCH_TOP_TEXT = 2
SEARCH_SUBWIDGETS = 3
SEARCH_BOTTOM_TEXT = 4

class MoultiCollapsible(Collapsible):
	"""
	Moulti-specific variant of Textual's Collapsible widget.
	"""

	def _watch_collapsed(self, collapsed: bool) -> None:
		"""Override so as not to scroll when expanding."""
		self._update_collapsed(collapsed)
		message = self.Collapsed(self) if self.collapsed else self.Expanded(self)
		self.post_message(message)

class CollapsibleStep(AbstractStep):
	"""
	This is the base class for all collapsible components end users may wish to
	add to Moulti.
	"""

	def __init__(self, id: str, **kwargs: Any): # pylint: disable=redefined-builtin
		self.collapsible = MoultiCollapsible(title=id)
		self.top_label = Static('', classes='top_text')
		self.bottom_label = Static('', classes='bottom_text')
		self.search_cursor = SEARCH_RESET

		self.top_text = ''
		self.bottom_text = ''

		super().__init__(id=id, **kwargs)

	def subcompose(self) -> ComposeResult:
		return []

	def compose(self) -> ComposeResult:
		with self.collapsible:
			yield self.top_label
			yield from self.subcompose()
			yield self.bottom_label

	def subsearch(self, _: TextSearch) -> bool:
		return False

	def search_part(self, search: TextSearch) -> tuple[bool, Widget|None]:
		if self.search_cursor == SEARCH_TITLE:
			found, text = self.search_label(self.title, search)
			self.collapsible.title = text.markup
			widget = self.query_one('CollapsibleTitle') if found else None
			return found, widget
		if self.search_cursor == SEARCH_TOP_TEXT:
			found, text = self.search_label(self.top_text, search)
			self.top_label.update(text)
			widget = self.top_label if found else None
			return found, widget
		if self.search_cursor == SEARCH_BOTTOM_TEXT:
			found, text = self.search_label(self.bottom_text, search)
			self.bottom_label.update(text)
			widget = self.bottom_label if found else None
			return found, widget
		if self.search_cursor == SEARCH_SUBWIDGETS:
			return self.subsearch(search), None
		return False, None

	def search(self, search: TextSearch) -> bool:
		found = False
		widget = None
		if search.reset():
			if self.search_cursor != SEARCH_RESET:
				self.search_part(search)
			self.search_cursor = SEARCH_RESET
			return False
		if search.next_result:
			parts = (SEARCH_TITLE, SEARCH_TOP_TEXT, SEARCH_SUBWIDGETS, SEARCH_BOTTOM_TEXT, SEARCH_RESET)
		else:
			parts = (SEARCH_BOTTOM_TEXT, SEARCH_SUBWIDGETS, SEARCH_TOP_TEXT, SEARCH_TITLE, SEARCH_RESET)
		if self.search_cursor == SEARCH_RESET:
			self.search_cursor = parts[0]
		while self.search_cursor != SEARCH_RESET:
			found, widget = self.search_part(search)
			if found:
				break
			self.search_cursor = parts[parts.index(self.search_cursor) + 1]
		if found:
			# Expand this collapsible step if necessary:
			if self.search_cursor != SEARCH_TITLE:
				self.collapsible.collapsed = False
			# Grab the focus, except for subwidgets, which have to grab focus explicitly:
			if self.search_cursor != SEARCH_SUBWIDGETS:
				self.focus(False)
			# Scroll to make the highlighted result visible:
			if widget is not None:
				self.post_message(self.ScrollRequest(self, widget, center=True, animate=False))
		return found

	def on_mount(self) -> None:
		self.query_one('CollapsibleTitle').tooltip = f'Step id: {self.title_from_id()}'

	def focus(self, scroll_visible: bool = True) -> Self:
		self.query_one('CollapsibleTitle').focus(scroll_visible)
		return self

	async def on_click(self, event: Click) -> None:
		"""
		Steps are meant to be focusable but it is actually the CollapsibleTitle that holds the focus.
		"""
		if event.control and event.control.can_focus:
			return
		self.focus()

	def check_properties(self, kwargs: dict[str, str|int|bool]) -> None:
		self.check_markup_dict(kwargs, 'top_text', 'bottom_text')
		super().check_properties(kwargs)

	def update_properties(self, kwargs: dict[str, str|int|bool]) -> None:
		if 'collapsed' in kwargs:
			self.collapsible.collapsed = bool(kwargs['collapsed'])
		if 'top_text' in kwargs:
			self.top_text = str(kwargs['top_text'])
			self.top_label.update(self.top_text)
		self.top_label.styles.display = 'block' if self.top_text else 'none'
		if 'bottom_text' in kwargs:
			self.bottom_text = str(kwargs['bottom_text'])
			self.bottom_label.update(self.bottom_text)
		self.bottom_label.styles.display = 'block' if self.bottom_text else 'none'
		super().update_properties(kwargs)
		self.collapsible.title = self.title

	def export_properties(self) -> dict[str, Any]:
		prop = super().export_properties()
		prop['collapsed'] = self.collapsible.collapsed
		prop['top_text'] = self.top_text
		prop['bottom_text'] = self.bottom_text
		return prop

	DEFAULT_CSS = AbstractStep.DEFAULT_CSS + """
	CollapsibleStep {
		/* Compact design: no padding, no margins, no borders: */
		& Collapsible {
			/* Inherit the parent background instead of altering it via $boost: */
			background: initial;
			padding: 0;
			margin: 0;
			border: none;
		}
		& CollapsibleTitle {
			padding: 0;
			color: $text;
			width: 100%;
		}
		& CollapsibleTitle:focus {
			background: initial;
			color: $text;
			text-style: none;
		}
		& CollapsibleTitle:hover {
			background: $foreground 80%;
			color: $text;
			text-style: none;
		}
		/* Collapsible contents: */
		& Contents {
			/* Leave some space on each side of the contents: */
			padding-left: 2 !important;
			padding-right: 1 !important;
		}
	}
	"""



import re
import sys
import subprocess
from os import getpid
from io import StringIO
from typing import Any, Generator
from unidiff import PatchSet # type: ignore
from .pipeline import pipeline
from .environ import env

ANSI_GREEN = '\x1b[0;32m' # ANSI green bold is hard to read on light backgrounds
ANSI_RED_BOLD = '\x1b[1;31m'
ANSI_RESET = '\x1b[0m'
DIFF_FIRST_LINE_RE = r'^(?:diff|index|\-\-\-|\+\+\+|Binary files)\s'
ENCODING = env('MOULTI_DIFF_ENCODING') or 'utf-8'

def separate_header_and_data(filedesc: Any) -> tuple[str, str]:
	"""
	unidiff discards all lines above the first file/hunk.
	"""
	header = []
	diff_data = ''
	while line := filedesc.readline():
		if re.search(DIFF_FIRST_LINE_RE, line):
			diff_data = line + filedesc.read()
			break
		header.append(line)
	return ''.join(header), diff_data

def colorize_hunk(hunk: str, strip_first_line: bool = True) -> str:
	lines = []
	for line_num, line in enumerate(hunk.split('\n')):
		if not line_num and strip_first_line:
			continue
		color = None
		if line.startswith('+'):
			color = ANSI_GREEN
		elif line.startswith('-'):
			color = ANSI_RED_BOLD
		if color:
			line = color + line + ANSI_RESET
		lines.append(line)
	return '\n'.join(lines)

def commands(title: str, header: str, diff: PatchSet) -> Generator:
	"""
	Yield pipeline()-compatible triplets.
	"""
	if env('MOULTI_DIFF_NO_TITLE') is None:
		yield None, {'command': 'set', 'title': title}, None

	delta_lines: list[str] = []
	if env('MOULTI_DIFF_NO_DELTA') is None:
		try:
			delta = ['delta', '--color-only']
			diff_s = str(diff)
			delta_output = subprocess.check_output(delta, input=diff_s, encoding=ENCODING, errors='surrogateescape')
			delta_lines = delta_output.splitlines()
		except Exception:
			pass

	print_steps = bool(env('MOULTI_DIFF_VERBOSE'))
	pid = getpid()
	counter = {'step': 1}
	def step(cmd: str, title: str, **kwargs: Any) -> tuple[str, dict, None]:
		step_id = f'diff_{pid}_{counter["step"]}'
		counter['step'] += 1
		step = {'command': cmd, 'action': 'add', 'id': step_id, 'title': title, 'classes': 'diff', **kwargs}
		if print_steps:
			print(f'{cmd} {step_id}')
		return step_id, step, None

	if header:
		yield step('step', 'Header', text=header, bottom_text=' ', collapsed=False)
	for file in diff:
		# One divider per file:
		step_id, divider, fd = step('divider', file.path)

		if file.is_rename:
			source_file = file.source_file
			if file.source_file.startswith('a/') and file.target_file.startswith('b/'):
				source_file = source_file[2:]
			divider['title'] = f'renamed: {source_file} -> {file.path}'
			divider['classes'] += ' warning'
		elif file.is_removed_file:
			divider['title'] = f'removed: {file.path}'
			divider['classes'] += ' error'
		elif file.is_added_file:
			divider['title'] = f'added: {file.path}'
			divider['classes'] += ' success'

		if file.is_binary_file:
			divider['title'] += ' (binary)'

		yield step_id, divider, fd

		for hunk in file:
			# One step per hunk:
			source = f'{hunk.source_start},{hunk.source_length}'
			target = f'{hunk.target_start},{hunk.target_length}'
			# Green and pink colors with a 7.3 contrast ratio, assuming background color #004578:
			stats = f'[#00FF66]+{hunk.added}[/][#FFD0EB]-{hunk.removed}[/]'
			title = f' [yellow1]@@ {source} {target} @@[/] {stats} [gray]{hunk.section_header}[/]'
			if delta_lines:
				hunk_first_line = hunk[0].diff_line_no
				hunk_last_line = hunk[-1].diff_line_no
				text = '\n'.join(delta_lines[hunk_first_line-1:hunk_last_line])
			else:
				text = colorize_hunk(str(hunk), True)
			yield step('step', title, text=text, bottom_text=' ', collapsed=False, max_height=0)

def handle_diff_from_file_descriptor(title: str, filedesc: Any) -> None:
	header, diff_data = separate_header_and_data(filedesc)
	diff = PatchSet.from_string(diff_data)
	pipeline(commands(title, header, diff))

def handle_diff_from_string(title: str, data: str) -> None:
	handle_diff_from_file_descriptor(title, StringIO(data))

def diff_parse(args: dict) -> None:
	filepath = args['diff_filepath']
	with open(filepath, 'r', encoding=ENCODING, errors='surrogateescape') as filedesc:
		title = f'{filepath}'
		handle_diff_from_file_descriptor(title, filedesc)

def diff_run(args: dict) -> None:
	command = args['command']
	res = subprocess.run(command, check=False, stdout=subprocess.PIPE, encoding=ENCODING, errors='surrogateescape')
	if res.stdout:
		title = ' '.join(command)
		handle_diff_from_string(title, res.stdout)
	if res.returncode and not res.stdout:
		sys.stderr.write(f'diff_run: {command} produced no output and exited with status {res.returncode}\n')
		sys.exit(res.returncode)
	sys.exit(0)



from textual.app import App
from pyperclip import copy as pyperclip_copy # type: ignore
from pyperclip import PyperclipException
from .environ import env

COPY_POLICY = env('MOULTI_CLIPBOARD_COPY')

class ClipboardException(Exception):
	pass

def osc_52_copy(app: App, data: str) -> None:
	app.copy_to_clipboard(data)
	# Alas, it is not possible to determine whether this method actually succeeded.

def copy(app: App, data: str) -> None:
	if COPY_POLICY == 'terminal-osc-52':
		osc_52_copy(app, data)
	elif COPY_POLICY == 'pyperclip':
		pyperclip_copy(data)
	else:
		# Default policy: pyperclip with failover to OSC 52:
		try:
			pyperclip_copy(data)
		except PyperclipException as pyperclip_exc:
			try:
				osc_52_copy(app, data)
			except ClipboardException as osc_52_exc:
				# pylint: disable=raise-missing-from
				raise ClipboardException(f'{pyperclip_exc}\nAdditionally, {osc_52_exc}')



from .common import moulti_test
assert moulti_test

COMMAND = ['text-search.bash']

def test_search_bar(moulti_test):
	bad_regex, good_regex = r'[aeiou', r'[aeiou]'
	assert moulti_test(press=('/', *bad_regex)) # Regex + red pattern
	assert moulti_test(press=('/', 'ctrl+r', *bad_regex)) # Plain + white pattern
	assert moulti_test(press=('/', 'ctrl+r', 'ctrl+t', *good_regex)) # PLAIN + white pattern
	assert moulti_test(press=('/', 'ctrl+t', *good_regex)) # REGEX + white pattern

def test_plain_pattern_case_sensitive(moulti_test):
	assert moulti_test(command=COMMAND, press=('/', 'ctrl+r', *'sed', 'enter', 'n')) # second "sed"
	assert moulti_test(command=COMMAND, press=('/', 'ctrl+r', *'sed', 'enter', 'n', 'N')) # first "sed"

def test_plain_pattern_case_insensitive(moulti_test):
	assert moulti_test(command=COMMAND, press=('/', 'ctrl+r', 'ctrl+t', *'sed', 'enter')) # first "Sed"

def test_regex_pattern_case_sensitive(moulti_test):
	assert moulti_test(command=COMMAND, press=('/', *r'\b\w{3}\b', 'enter', 'n', 'n', 'n')) # first "vel"

def test_regex_pattern_case_insensitive(moulti_test):
	assert moulti_test(command=COMMAND, press=('/', *r'\blor[aeiou]m\b', 'ctrl+t', 'enter')) # first "Lorem"



from argparse import _SubParsersAction
from moulti.client import send_delete, send_to_moulti_and_handle_reply, send_no_none_to_moulti_and_handle_reply
from ..abstractstep.cli import add_abstractstep_options

COMMAND = 'divider'

def add_cli_arguments(subparsers: _SubParsersAction) -> None:
	# moulti divider
	divider_parser = subparsers.add_parser(COMMAND, help='create and manage dividers')
	divider_subparsers = divider_parser.add_subparsers(required=True)
	add_divider_commands(divider_subparsers)

def add_divider_commands(divider_subparsers: _SubParsersAction) -> None:
	# moulti divider add
	divider_add_parser = divider_subparsers.add_parser('add', help='add a new divider to Moulti')
	divider_add_parser.set_defaults(func=send_to_moulti_and_handle_reply, command=COMMAND, action='add')
	divider_add_parser.add_argument('id', type=str, help='unique identifier')
	add_abstractstep_options(divider_add_parser)

	# moulti divider update
	divider_update_parser = divider_subparsers.add_parser('update', help='update an existing Moulti divider')
	divider_update_parser.set_defaults(func=send_no_none_to_moulti_and_handle_reply, command=COMMAND, action='update')
	divider_update_parser.add_argument('id', type=str, help='unique identifier')
	add_abstractstep_options(divider_update_parser, none=True)

	# moulti divider delete
	divider_delete_parser = divider_subparsers.add_parser('delete', help='delete an existing Moulti divider')
	divider_delete_parser.set_defaults(func=send_delete, command=COMMAND, action='delete')
	divider_delete_parser.add_argument('id', type=str, nargs='+', help='unique identifier')



from typing import Any
from textual.app import ComposeResult
from textual.widgets import Button
from ..buttonquestion.tui import ButtonQuestion
from ..inputquestion.tui import InputQuestion

class Question(ButtonQuestion, InputQuestion):
	"""
	This widget represents an interactive question in a script, program or process.
	It combines the abilities of InputQuestion (single input field) and ButtonQuestion (buttons).
	"""

	def __init__(self, id: str, **kwargs: Any): # pylint: disable=redefined-builtin
		ButtonQuestion.__init__(self, id=id, **kwargs)
		InputQuestion.__init__(self, id=id, **kwargs)

	def compose_question(self) -> ComposeResult:
		yield from InputQuestion.compose_question(self)
		yield from ButtonQuestion.compose_question(self)

	def on_button_pressed(self, event: Button.Pressed) -> None:
		# Do not call ButtonQuestion.on_button_pressed():
		event.prevent_default()
		answer = event.button.name
		if answer is not None:
			answer = answer.replace('{input}', self.input.value)
			self.got_answer(answer)

	def check_properties(self, kwargs: dict[str, Any]) -> None:
		ButtonQuestion.check_properties(self, kwargs)
		InputQuestion.check_properties(self, kwargs)

	def export_properties(self) -> dict[str, Any]:
		return {
			**InputQuestion.export_properties(self),
			**ButtonQuestion.export_properties(self)
		}

MoultiWidgetClass = Question



from .common import moulti_test, moulti_test_steps
assert moulti_test

def test_progressbar(moulti_test):
	"""
	Run commands taken from the documentation's "Progress bar" section.
	"""
	moulti_test_steps(moulti_test, 'progressbar.bash', 4)



from .common import Beacon, moulti_test, moulti_test_steps, wait_script, wait_script_start, wait_script_end
assert moulti_test

def test_programmatic_scrolling(moulti_test):
	"""
	Run most commands from the documentation's "Programmatically scrolling through steps" section.
	"""
	moulti_test_steps(moulti_test, 'prog-scrolling-through-steps.bash', 7)

def test_lock_scroll(moulti_test):
	# No scroll lock: do not wait, let it scroll:
	assert moulti_test(command=['manual-scrolling-through-steps.bash', '0'])
	# Scroll lock:
	async def scroll_lock(pilot):
		await pilot.press('l') # enable scroll lock
		await wait_script(pilot)
	assert moulti_test(command=['manual-scrolling-through-steps.bash'], run_before=scroll_lock)

def test_auto_scrolling(moulti_test, monkeypatch):
	beacon = Beacon(monkeypatch)
	async def interrupt_auto_scroll(pilot):
		await wait_script_start(pilot)
		await beacon.wait(pilot, '-output-started')
		await pilot.pause(0.5)
		await pilot.press('tab', 'tab', 'home') # focus the first step, focus its log, hit the 'home' key
		await wait_script_end(pilot)
		# Wait until the first step has finished scrolling back up to the top:
		log_widget = pilot.app.steps_container.children[0].log_widget
		interval = 0.1
		max_wait = int(45 / interval)
		while max_wait and log_widget.scroll_y:
			await pilot.pause(interval)
			max_wait -= 1
	assert moulti_test(command=['scrolling-inside-steps.bash'], run_before=interrupt_auto_scroll)



from textual.app import ComposeResult
from textual.containers import Grid
from textual.events import Key
from textual.message import Message
from textual.screen import ModalScreen
from textual.widgets import Button, Label

class QuitDialog(ModalScreen):
	"""
	Modal dialog asking the user whether they actually want to quit Moulti,
	along with an optional extra message detailing why confirmation is required.
	Mostly based on https://textual.textualize.io/guide/screens/#modal-screens
	"""
	def __init__(self, message: str = '', allow_terminate: bool = False) -> None:
		self.message = message
		self.allow_terminate = allow_terminate
		self.quit_request_count = 1
		super().__init__()

	def compose(self) -> ComposeResult:
		base_message = 'Are you sure you want to quit Moulti?'
		if self.message:
			base_message += '\n' + self.message
		with Grid(id='quit_dialog_grid', classes='terminate' if self.allow_terminate else ''):
			yield Label(base_message, id='quit_dialog_question')
			yield Button('Do not quit (esc)', variant='primary', id='do_not_quit')
			if self.allow_terminate:
				yield Button('Quit and leave the process running in the background (l)', variant='warning', id='quit')
				yield Button('Terminate the process and quit (t)', variant='error', id='terminate_and_quit')
			else:
				yield Button('Quit (l)', variant='warning', id='quit')
			yield Label('', id='extra_info')

	class ExitRequest(Message):
		def __init__(self, exit_first_policy: str):
			self.exit_first_policy = exit_first_policy
			super().__init__()

	def on_button_pressed(self, event: Button.Pressed) -> None:
		if event.button.id == 'do_not_quit':
			self.app.pop_screen()
		else:
			policy = 'terminate' if event.button.id == 'terminate_and_quit' else ''
			self.exit(policy)

	def on_key(self, event: Key) -> None:
		buttons = {
			'escape': '#do_not_quit',
			'l': '#quit',
		}
		if self.allow_terminate:
			buttons['t'] = '#terminate_and_quit'
		if button_id := buttons.get(event.key.lower()):
			self.query_one(button_id, Button).action_press()

	def exit(self, policy: str) -> None:
		self.post_message(self.ExitRequest(policy))

	def new_quit_request(self) -> None:
		self.quit_request_count += 1
		if self.allow_terminate:
			if self.quit_request_count == 2:
				self.query_one('#quit').focus()
			elif self.quit_request_count == 3:
				x = self.query_one('#extra_info', Label)
				x.update('Hitting Ctrl+q again will terminate the process and quit.')
				self.query_one('#terminate_and_quit').focus()
			elif self.quit_request_count > 3:
				self.exit('terminate')
		else:
			if self.quit_request_count == 2:
				x = self.query_one('#extra_info', Label)
				x.update('Hitting Ctrl+q again will quit.')
				self.query_one('#quit').focus()
			elif self.quit_request_count == 3:
				self.exit('leave')

	DEFAULT_CSS = """
		QuitDialog {
			align: center middle;
			background: $surface 60%;
			& Button {
				width: 100%;
			}
		}
		#do_not_quit {
			background: $secondary;
		}

		#quit_dialog_grid {
			grid-size: 1 4;
			grid-gutter: 1 2;
			grid-rows: 1fr 1fr 1fr 1;
			padding: 0 1;
			width: 65;
			height: 15;
			border: thick $surface;
			background: $background;
			&.terminate { /* 1 extra button */
				grid-size: 1 5;
				grid-rows: 1fr 1fr 1fr 1fr 1;
				height: 19;
			}
		}

		#quit_dialog_question, #extra_info {
			height: 1fr;
			width: 100%;
			content-align: center middle;
		}
	"""



from pathlib import Path
from moulti.app import Moulti
from .common import moulti_test
assert moulti_test

def test_step_classes(moulti_test, monkeypatch):
	css_path = Path('tests/styles/customstate.css').absolute()
	monkeypatch.setattr(Moulti, 'CSS_PATH', str(css_path))
	assert moulti_test(command=['step-classes.bash'])



from typing import Any
from textual.app import ComposeResult
from textual.widgets import Input
from ..abstractquestion.tui import AbstractQuestion

class InputQuestion(AbstractQuestion):
	"""
	This widget represents an interactive question in a script, program or process.
	"""

	BINDINGS = [
		("c", "to_clipboard(False)", "Copy"),
		("w", "to_clipboard(True)", "With question"),
	]

	def __init__(self, id: str, **kwargs: Any): # pylint: disable=redefined-builtin
		super().__init__(id=id, **kwargs)
		self.input = Input()

	def compose_question(self) -> ComposeResult:
		yield self.input

	def on_input_submitted(self, event: Input.Submitted) -> None:
		answer = event.input.value
		if answer is not None:
			self.got_answer(answer)

	def update_properties(self, kwargs: dict[str, str|int|bool]) -> None:
		super().update_properties(kwargs)
		if 'placeholder' in kwargs:
			self.input.placeholder = str(kwargs['placeholder'])
		if 'value' in kwargs:
			self.input.value = str(kwargs['value'])
		if 'password' in kwargs:
			self.input.password = bool(kwargs['password'])
		if 'max_length' in kwargs:
			try:
				max_length = int(kwargs['max_length'])
				self.input.max_length = None if max_length == 0 else max_length
			except (ValueError, TypeError):
				pass
		if 'restrict' in kwargs:
			self.input.restrict = None if not kwargs['restrict'] else str(kwargs['restrict'])

	def export_properties(self) -> dict[str, Any]:
		prop = super().export_properties()
		prop['placeholder'] = str(self.input.placeholder)
		# Do NOT export the value if the input field is in "password mode":
		prop['value'] = str(self.input.value) if not self.input.password else '*'*len(self.input.value)
		prop['password'] = bool(self.input.password)
		prop['max_length'] = 0 if self.input.max_length is None else self.input.max_length
		prop['restrict'] = '' if self.input.restrict is None else self.input.restrict
		return prop

	@AbstractQuestion.copy_to_clipboard
	def action_to_clipboard(self, with_question: bool = False) -> tuple[bool, str, str]:
		answer = self.input.value
		data = f'Q: {self.question}\nA: {answer}\n' if with_question else answer
		return True, data, ''

MoultiWidgetClass = InputQuestion



import errno
import os
import selectors
from queue import Queue, Empty
from stat import S_ISREG
from time import monotonic_ns
from typing import Any, Callable, Iterable, Sequence
from textual import work
from textual.app import ComposeResult
from textual.strip import Strip
from textual.worker import get_current_worker
from rich.ansi import re_ansi
from rich.text import Text
from rich.cells import get_character_cell_size
from moulti.helpers import ANSI_ESCAPE_SEQUENCE_BYTES, ANSI_RESET_SEQUENCES_BYTES, TAB_SPACES_BYTES, clean_selector
from moulti.search import TextSearch
from . import MOULTI_PASS_DEFAULT_READ_SIZE
from ..collapsiblestep.tui import CollapsibleStep, SEARCH_SUBWIDGETS
from ..moultilog import MoultiLog

class ThrottledAppender:
	"""
	This class uses Textual's call_from_thread() to schedule calls to "step.append_lines()" inside Textual's asyncio
	event loop in a thread-safe manner. These calls are throttled so as not to flood the event loop.
	"""
	def __init__(self, step: 'Step', delay_ms: int):
		self.step = step
		self.call_from_thread = step.app.call_from_thread
		self.delay_ns = delay_ms * 1_000_000
		self.lines: list[str|Text] = []
		self.max_cell_len = -1
		self.last_append = 0

	def new_data(self, lines: Iterable[str|Text], max_cell_len: int, force: bool = False) -> None:
		self.lines.extend(lines)
		if max_cell_len > self.max_cell_len: # pylint: disable=consider-using-max-builtin
			self.max_cell_len = max_cell_len
		self.append(force)

	def append(self, force: bool = False) ->  None:
		if self.lines:
			now = monotonic_ns()
			if force or (now - self.last_append >= self.delay_ns):
				self.call_from_thread(self.step.append_lines, self.lines, self.max_cell_len)
				self.lines = []
				self.max_cell_len = -1
				self.last_append = now

class Step(CollapsibleStep):
	"""
	This widget represents a step in a script, program or process.
	Visually speaking, it is essentially a collapsible text area surrounded
	with optional text lines.
	"""

	BINDINGS = [
		("c", "to_clipboard(False)", "Copy"),
		("w", "to_clipboard(True)", "With colors"),
	]

	def __init__(self, id: str, **kwargs: str|int|bool): # pylint: disable=redefined-builtin
		self.log_widget = MoultiLog()

		self.min_height = 1
		self.max_height = 25

		super().__init__(id=id, **kwargs)

		self.color = ''

	def cli_action_append(self, kwargs: dict[str, Any], helpers: dict[str, Any]) -> tuple:
		if 'text' not in kwargs:
			helpers['reply'](done=False, error='missing text for append operation')
			return ()
		return self.append, '\n'.join(kwargs['text'])

	def cli_action_clear(self, _kwargs: dict[str, str|int|bool], _helpers: dict[str, Any]) -> tuple:
		return self.clear, # pylint: disable=trailing-comma-tuple

	def subcompose(self) -> ComposeResult:
		yield self.log_widget

	def subsearch(self, search: TextSearch) -> bool:
		if found := self.log_widget.search(search):
			# We found a match inside the log widget:
			if self.collapsible.collapsed:
				self.ensure_expanded()
			self.show_search_highlight()
		return found

	def search_maximized(self, search: TextSearch) -> bool:
		if self.screen.maximized != self.log_widget:
			return False
		if found := self.log_widget.search(search):
			# A match was found inside the maximized widget:
			# 1 - ensure no other part is highlighted:
			if self.search_cursor != SEARCH_SUBWIDGETS:
				self.search(TextSearch.make_reset())
			# 2 - update the search cursor so as to resume search from it:
			self.search_cursor = SEARCH_SUBWIDGETS
			self.log_widget.scroll_to_search_highlight()
		return found

	def ensure_expanded(self, expanded: bool = True) -> None:
		self.collapsible.collapsed = not expanded
		# Collapsible rely on CSS: this step is not actually expanded until CSS is refreshed:
		self.app.refresh_css(animate=False)

	def show_search_highlight(self) -> None:
		region_spacing = self.log_widget.scroll_to_search_highlight()
		# To make scrolling more convenient, give the focus to the log widget if and only if it can be scrolled:
		if self.log_widget.max_scroll_x or self.log_widget.max_scroll_y:
			self.log_widget.focus(False)
		else: # otherwise, focus the step, thus enabling scrolling the StepContainer:
			self.focus(False)
		if region_spacing is not None:
			region, _ = region_spacing # relative to the MoultiLog widget
			log_widget_offset = self.log_widget.region.offset - self.region.offset
			region = region.translate(log_widget_offset) # relative to this Step widget
			self.post_message(self.ScrollRequest(self, region, center=True, animate=False))

	def update_properties(self, kwargs: dict[str, str|int|bool]) -> None:
		super().update_properties(kwargs)
		if 'auto_scroll' in kwargs:
			self.log_widget.auto_scroll = self.log_widget.default_auto_scroll = bool(kwargs['auto_scroll'])
		if 'text' in kwargs:
			self.clear()
			self.append(str(kwargs['text']))
		if 'min_height' in kwargs:
			self.min_height = int(kwargs['min_height'])
		if 'max_height' in kwargs:
			self.max_height = int(kwargs['max_height'])
		self.log_widget.styles.min_height = self.min_height
		self.log_widget.set_max_height(self.max_height)

	def export_properties(self) -> dict[str, Any]:
		prop = super().export_properties()
		prop['auto_scroll'] = self.log_widget.default_auto_scroll
		prop['min_height'] = self.min_height
		prop['max_height'] = self.max_height
		return prop

	def save(self, opener: Callable[[str, int], int], filename: str, extra_properties: dict[str, Any]) -> None:
		super().save(opener, filename, extra_properties)
		filename = filename + '.contents.log'
		with open(filename, 'w', encoding='utf-8', errors='surrogateescape', opener=opener) as contents_filedesc:
			self.log_widget.to_file(contents_filedesc)

	@CollapsibleStep.copy_to_clipboard
	def action_to_clipboard(self, keep_styles: bool = True) -> tuple[bool, str, str]:
		lines = list(self.log_widget.to_lines(keep_styles))
		lines_count = len(lines)
		lines.append('') # add an extra \n
		data = '\n'.join(lines)
		return True, data, f'copied {lines_count} lines, {len(data)} characters to clipboard'

	def clear(self) -> None:
		self.log_widget.clear()
		self.color = ''

	def append(self, text: str) -> None:
		if not text:
			return
		if text[-1] != '\n':
			text += '\n'
		# This looks silly, but our main ingestion function takes bytes, not strings:
		text_bytes = text.encode('utf-8')
		lines, max_cell_len, _, _ = Step.bytes_to_lines(text_bytes, b'', 'utf-8')
		self.append_lines(lines, max_cell_len)

	def append_lines(self, lines: Iterable[str|Text|Strip], max_cell_len: int) -> None:
		self.log_widget.write_lines(lines, max_cell_len)
		self.activity()

	@classmethod
	def last_character_color(cls, text: Text) -> str:
		# If the last span (if any) covers the last character...
		if text.spans and text.spans[-1].end == len(text):
			# ... return the ANSI escape code for its color:
			style = text.spans[-1].style
			assert not isinstance(style, str) # prevent calling render() on str
			return style.render('_').split('_')[0]
		return ''

	def cli_action_pass(self, kwargs: dict[str, str|int|bool], helpers: dict[str, Any]) -> tuple:
		if not helpers['file_descriptors']:
			helpers['reply'](done=False, error='missing file descriptor for pass operation')
			return ()
		# Set up a queue between two workers:
		# - one that reads data from the file descriptor and replies to the client;
		# - one that appends lines to the step.
		queue: Queue = Queue()
		self.append_from_queue(queue, helpers)
		self.append_from_file_descriptor_to_queue(queue, kwargs, helpers)
		return ()


	def watch_read_events(self, selector: selectors.BaseSelector, file_descriptor: int) -> Callable[[], bool]:
		try:
			# On Linux, epoll() does not support regular files and returns EPERM (register() raises an exception).
			# On macOS, it is possible to watch read events for a regular file... but that induces problems when the
			# file is (and remains) empty, because no events are generated.
			# In the end, it is simpler to NOT watch regular files:
			if S_ISREG(os.fstat(file_descriptor).st_mode):
				raise OSError('attempt to watch read events for a regular file')
			selector.register(file_descriptor, selectors.EVENT_READ)
			return lambda: bool(selector.select(0.5))
		except Exception:
			return lambda: True

	@work(thread=True, group='step-ingestion', name='fd-to-queue')
	async def append_from_file_descriptor_to_queue(
		self,
		queue: Queue,
		kwargs: dict[str, Any],
		helpers: dict[str, Any],
	) -> None:
		current_worker = get_current_worker()
		error = None
		original_file_descriptor = None
		output_selector = None
		try:
			original_file_descriptor = helpers['file_descriptors'][0]
			if isinstance(original_file_descriptor, int):
				file_descriptor = original_file_descriptor
			else:
				file_descriptor = original_file_descriptor.fileno()
			# Although this thread is dedicated to reading from a single file descriptor, it should:
			# - not block indefinitely on a read()
			# - check is_cancelled at regular intervals
			# This ensures Moulti exits correctly.
			# To this end, use non-blocking mode along with a select()-like interface:
			os.set_blocking(file_descriptor, False)
			output_selector = selectors.DefaultSelector()
			can_read = self.watch_read_events(output_selector, file_descriptor)
			# Read binary data from the given file descriptor using a FileIO (raw binary stream whose methods only make
			# one system call):
			with os.fdopen(file_descriptor, mode='rb', buffering=0) as binary_stream:
				read_size = int(kwargs.get('read_size', MOULTI_PASS_DEFAULT_READ_SIZE))
				if read_size <= 0:
					read_size = MOULTI_PASS_DEFAULT_READ_SIZE
				end_of_file = False
				# Outer loop: check for worker cancellation and rely on select() to detect activity:
				while not current_worker.is_cancelled and not end_of_file:
					if can_read():
						# Inner loop: read data until it would block, keep checking for worker cancellation:
						# "If the object is in non-blocking mode and no bytes are available, None is returned."
						while (data := binary_stream.read(read_size)) is not None:
							# "If 0 bytes are returned, and size was not 0, this indicates end of file."
							if not data:
								end_of_file = True
								break
							queue.put_nowait(data)
							if current_worker.is_cancelled:
								break
				queue.put_nowait(None)
		except Exception as exc:
			error = str(exc)
			helpers['debug'](f'pass: {error}')
		finally:
			try:
				clean_selector(output_selector, close_fds=False, close=True)
				if original_file_descriptor is not None and not isinstance(original_file_descriptor, int):
					# The file descriptor passed to this method is an object, and its fileno() was extracted to
					# instantiate a second fileobject (binary_stream), leading to TWO objects referencing the same
					# underlying file descriptor, which was typically closed upon exiting the with statement above.
					# Ideally, we should prevent the first fileobject from calling close() (or any equivalent private
					# method) because that double-close could strike another file descriptor opened in the meantime,
					# leading to one hell of a troubleshooting session.
					# This is particularly relevant whenever the Python process keeps running after the Moulti App has
					# finished running and garbage collection is not immediate.
					# In practice, we can simply call close() right now and handle the errno=EBADF that typically
					# ensues.
					try:
						original_file_descriptor.close()
					except OSError as ose:
						if ose.errno != errno.EBADF: # Bad file descriptor
							raise
			except Exception as exc:
				error = str(exc)
				helpers['debug'](f'pass: cleanup: unexpected error: {error}')
		helpers['reply'](done=error is None, error=error)

	@classmethod
	def bytes_to_lines(
		cls,
		data: bytes,
		color: bytes = b'',
		encoding: str = 'utf-8'
	) -> tuple[Sequence[str|Text], int, bytes, bytes]:
		lines: list[str|Text] = []
		max_cell_len: int = -1
		leftover: bytes = b''
		if data:
			# Deal only with lines ending with a line feed (lf, \n) and return leftover bytes:
			if data[-1] != b'\n':
				last_lf_index = data.rfind(b'\n')
				leftover = data[last_lf_index+1:]
				data = data[:last_lf_index]
			if data:
				# At this stage, all lines end with \n.
				for line_bytes in data.split(b'\n'):
					# Expand tabs into spaces: both Rich and Textual provide frighteningly
					# complex methods to do it the right way. But Text.expand_tabs()
					# dramatically and negatively affects performance.
					line_bytes = line_bytes.replace(b'\t', TAB_SPACES_BYTES)

					if ANSI_ESCAPE_SEQUENCE_BYTES in line_bytes:
						if Step.ends_with_ansi_reset(line_bytes):
							# Simple case: the last ANSI sequence in the line resets all styles:
							line_bytes = color + line_bytes
							line_str = line_bytes.decode(encoding, errors='surrogateescape')
							line_plain = re_ansi.sub('', line_str)
							lines.append(line_str)
							color = b''
						else:
							# The trailing underscore will determine the next line's color:
							line_bytes = color + line_bytes + b'_'
							line_str = line_bytes.decode(encoding, errors='surrogateescape')
							# ANSI sequences can be tricky: better call Text.from_ansi() even if it is expensive:
							line_text = Text.from_ansi(line_str)
							line_plain = line_text.plain
							color = cls.last_character_color(line_text).encode('ascii') # analyse the trailing underscore
							line_text.right_crop(1) # remove the trailing underscore
							lines.append(line_text)
					else:
						line_str = line_bytes.decode(encoding, errors='surrogateescape')
						line_plain = line_str
						if color:
							line_str = color.decode('ascii') + line_str
						lines.append(line_str)

					max_cell_len = Step.update_max_cell_len(max_cell_len, line_plain)
		return lines, max_cell_len, leftover, color

	@classmethod
	def ends_with_ansi_reset(cls, line: bytes) -> bool:
		index = line.rfind(ANSI_ESCAPE_SEQUENCE_BYTES)
		rest_of_line = line[index:]
		for reset_seq in ANSI_RESET_SEQUENCES_BYTES:
			if rest_of_line.startswith(reset_seq):
				return True
		return False


	@classmethod
	def update_max_cell_len(cls, max_cell_len: int, line_plain: str) -> int:
		line_plain_char_len = len(line_plain)
		# Worst-case: each character takes two cells:
		if (2 * line_plain_char_len) > max_cell_len:
			current_cell_len = 0
			for position, char in enumerate(line_plain):
				# Every 10 characters, evaluate the worst-case scenario again:
				if (position % 10) == 0 and position:
					if (current_cell_len + 2 * (line_plain_char_len - position)) <= max_cell_len:
						break
				current_cell_len += get_character_cell_size(char)
			else:
				if current_cell_len > max_cell_len: # pylint: disable=consider-using-max-builtin
					max_cell_len = current_cell_len
		return max_cell_len

	@work(thread=True, group='step-ingestion', name='queue-to-step')
	async def append_from_queue(self, queue: Queue, helpers: dict[str, Any]) -> None:
		current_worker = get_current_worker()
		self.prevent_deletion += 1
		color = b''
		try:
			throttling_ms = 25
			throttling_s = throttling_ms / 1000
			appender = ThrottledAppender(self, throttling_ms)
			buffer = []
			while True:
				if current_worker.is_cancelled:
					break
				try:
					data = queue.get(block=True, timeout=throttling_s)
					if data is not None:
						buffer.append(data)
						all_data = b''.join(buffer)
						lines, max_cell_len, leftover, color = Step.bytes_to_lines(all_data, color)
						if lines:
							appender.new_data(lines, max_cell_len)
						buffer.clear()
						if leftover:
							buffer.append(leftover)
					else: # Reached EOF: flush buffer and exit:
						if buffer:
							buffer.append(b'\n')
							all_data = b''.join(buffer)
							lines, max_cell_len, _, _ = Step.bytes_to_lines(all_data, color)
							appender.new_data(lines, max_cell_len)
						appender.append(True)
						break
				except Empty:
					# No data: there may be data left in the appender's buffer:
					appender.append(True)
		except Exception as exc:
			helpers['debug'](f'append_from_queue: {exc}')
		finally:
			self.prevent_deletion -= 1

	DEFAULT_CSS = CollapsibleStep.DEFAULT_CSS + """
	Step {
		& MoultiLog {
			scrollbar-corner-color: $step_default;
		}
		&.success {
			& MoultiLog { scrollbar-corner-color: $step_success; }
		}
		&.warning {
			& MoultiLog { scrollbar-corner-color: $step_warning; }
		}
		&.error {
			& MoultiLog { scrollbar-corner-color: $step_error; }
		}
		&.inactive {
			& MoultiLog { scrollbar-corner-color: $step_inactive; }
		}
	}
	"""
MoultiWidgetClass = Step



from typing import Any
from textual.app import ComposeResult
from textual.events import Resize
from textual.containers import Horizontal
from textual.reactive import Reactive
from textual.widgets import Button, Static
from ..abstractquestion.tui import AbstractQuestion
from .. import MoultiWidgetInvalidPropertyException

DEFAULT_BUTTONS = [
	['yes', 'success', 'Yes'],
	['no', 'error', 'No'],
	['cancel', 'primary', 'Cancel'],
]

DUMMY_BUTTON = Button()

def check_button_variant(variant: str) -> bool:
	try:
		Button.validate_variant(DUMMY_BUTTON, variant)
		return True
	# The actual exception is private:
	except Exception as exc:
		msg = f"invalid style '{variant}'. {exc}"
		raise MoultiWidgetInvalidPropertyException(msg) from exc

class ButtonQuestion(AbstractQuestion):
	"""
	This widget represents an interactive question in a script, program or process.
	"""

	BINDINGS = [
		("c", "to_clipboard(False)", "Copy"),
		("w", "to_clipboard(True)", "With question"),
	]

	known_width = Reactive(-1)

	def __init__(self, id: str, **kwargs: Any): # pylint: disable=redefined-builtin
		super().__init__(id=id, **kwargs)
		self.button_layout: list[int] = []
		self.button_sizes: list[int]|None = None
		self.regenerate_buttons()
		self.vertical = Static(classes='buttons')

	def regenerate_buttons(self) -> None:
		self.buttons = []
		disabled = self.answer is not None
		for value, classes, label in self.init_kwargs.get('button', DEFAULT_BUTTONS):
			self.buttons.append(Button(label, variant=classes, name=value, disabled=disabled))

	def compose_question(self) -> ComposeResult:
		with self.vertical:
			# Initial layout: all buttons in a single row:
			with Horizontal():
				yield from self.buttons

	def on_button_pressed(self, event: Button.Pressed) -> None:
		answer = event.button.name
		if answer is not None:
			self.got_answer(answer)

	def check_properties(self, kwargs: dict[str, Any]) -> None:
		super().check_properties(kwargs)
		for _, style, label in kwargs.get('button', []):
			check_button_variant(style)
			self.check_markup(label)

	def export_properties(self) -> dict[str, Any]:
		prop = super().export_properties()
		# Buttons cannot be updated so we can simply export the value found in init_kwargs:
		prop['button'] = self.init_kwargs.get('button', DEFAULT_BUTTONS)
		return prop

	@AbstractQuestion.copy_to_clipboard
	def action_to_clipboard(self, with_question: bool = False) -> tuple[bool, str, str]:
		if self.answer is None:
			return False, '', 'Answer not available yet'
		data = f'Q: {self.question}\nA: {self.answer}\n' if with_question else self.answer
		return True, data, ''

	def measure_buttons(self) -> None:
		if self.button_sizes is None:
			# Harvest current button sizes:
			self.button_sizes = [button.size.width + button.styles.margin.right for button in self.buttons]

	def compute_button_layout(self, width: int = -1) -> list[int]:
		"""
		Return a list of integers representing how many buttons each row should contain.
		"""
		width = self.query('Contents').first().size.width if width == -1 else width
		width -= self.buttons[0].styles.margin.left
		self.measure_buttons()
		rows = []
		row = 0
		row_width = width
		assert self.button_sizes
		for button_width in self.button_sizes:
			if button_width > row_width and row:
				rows.append(row)
				row = 0
				row_width = width
			row += 1
			row_width -= button_width
		rows.append(row)
		width = self.query('Contents').first().size.width
		return rows

	async def apply_button_layout(self, new_layout: list[int]) -> None:
		"""
		Given a list of integers representing how many buttons each row should contain, achieve that state by:
		- adding new Horizontal containers if needed
		- moving buttons
		- removing unnecessary Horizontal containers, if any
		"""
		horizontals = list(self.vertical.query('Horizontal').results())
		current_button_map: list[int] = []
		for horizontal_index, horizontal in enumerate(horizontals):
			current_button_map.extend(horizontal_index for _ in horizontal.query('Button').results())
		current_horizontal_count = max(current_button_map) + 1

		new_button_map: list[int] = []
		for horizontal_index, button_count in enumerate(new_layout):
			new_button_map.extend(horizontal_index for _ in range(button_count))
		new_horizontal_count = len(new_layout)

		buttons_to_move: dict[tuple[int, int], list[int]] = {}
		for button_index, current_horizontal_index in enumerate(current_button_map):
			new_horizontal_index = new_button_map[button_index]
			if current_horizontal_index != new_horizontal_index:
				move_direction = new_horizontal_index - current_horizontal_index
				move_direction //= abs(move_direction)
				buttons_to_move.setdefault((new_horizontal_index, move_direction), []).append(button_index)

		if not buttons_to_move:
			return

		# Add extra Horizontal containers if needed:
		if new_horizontal_count > current_horizontal_count:
			new_horizontals = list(Horizontal() for _ in range(current_horizontal_count, new_horizontal_count))
			horizontals.extend(new_horizontals)
			self.vertical.mount_all(new_horizontals)

		# Move buttons:
		disabled = self.answer is not None
		buttons_reference = self.init_kwargs.get('button', DEFAULT_BUTTONS)
		for movement, button_indices in buttons_to_move.items():
			fresh_buttons = []
			for button_index in button_indices:
				self.buttons[button_index].remove()
				button_value, button_classes, button_label = buttons_reference[button_index]
				fresh_button = Button(button_label, variant=button_classes, name=button_value, disabled=disabled)
				self.buttons[button_index] = fresh_button
				fresh_buttons.append(fresh_button)
			target_horizontal_index, direction = movement
			if direction > 0:
				horizontals[target_horizontal_index].mount_all(fresh_buttons, before=0)
			else:
				horizontals[target_horizontal_index].mount_all(fresh_buttons)

		# Remove unnecessary Horizontal containers, if any:
		if new_horizontal_count < current_horizontal_count:
			for horizontal_index in range(new_horizontal_count, current_horizontal_count):
				horizontals[horizontal_index].remove()

	async def watch_known_width(self, _old_width: int, _new_width: int) -> None:
		"""
		Compute and apply a new layout for buttons whenever the width changes.
		"""
		# Do nothing if we ignore our width:
		if self.known_width == -1:
			return
		# Do not bother for zero or one button:
		if len(self.buttons) <= 1:
			return
		# Do not bother if this widget is collapsed:
		if self.collapsible.collapsed:
			# Discard the known width to force a change when this widget is expanded again:
			self.known_width = -1
			return
		# Recompute the layout:
		new_layout = self.compute_button_layout()
		# Do not bother if the layout did not change:
		if new_layout == self.button_layout:
			return
		self.button_layout = new_layout
		await self.apply_button_layout(new_layout)

	def on_resize(self, event: Resize) -> None:
		if self.known_width != event.size.width:
			self.known_width = event.size.width

	DEFAULT_CSS = """
		Static.buttons {
			layout: vertical;
			&> Horizontal {
				align: center middle;
				height: 5; /* Button height + margin */
				&> Button { margin-right: 2; margin-left: 2; margin-top: 1; margin-bottom: 1; }
			}
		}
	"""
MoultiWidgetClass = ButtonQuestion



import json
from operator import attrgetter
from typing import Any, Callable
from rich.text import Text
from textual.geometry import Region
from textual.message import Message
from textual.widget import Widget
from textual.widgets import Static
from moulti.clipboard import copy
from moulti.search import MatchSpan, TextSearch

class AbstractStep(Static):
	"""
	This is the base class for all components end users may wish to add to Moulti.
	"""

	class StepActivity(Message):
		def __init__(self, step: 'AbstractStep', *args: Any, **kwargs: Any):
			self.step = step
			super().__init__(*args, **kwargs)

	class ScrollRequest(Message):
		def __init__(self, step: 'AbstractStep', target: Widget|Region|None = None, **scroll_kwargs: Any):
			self.step = step
			self.target = target
			self.scroll_kwargs = scroll_kwargs
			super().__init__()

	def __init__(self, id: str, **kwargs: Any): # pylint: disable=redefined-builtin
		self.title = id

		# This attribute is meant to prevent deletion of a step while content
		# is being appended to it:
		self.prevent_deletion = 0

		# This attribute is used by the search_label() method.
		self.label_search_cursor: MatchSpan = None

		self.scroll_on_activity: bool|int = False
		self.check_properties(kwargs)
		self.init_kwargs = kwargs

		step_classes = str(kwargs.get('classes', ''))
		super().__init__(id='step_' + id, classes=step_classes)

	def on_mount(self) -> None:
		self.update_properties(self.init_kwargs)

	def title_from_id(self) -> str:
		self_id = str(self.id)
		if '_' not in self_id:
			return self_id
		return self_id.split('_', 1)[1]

	def index(self) -> int:
		if self.parent is not None:
			return sorted(self.parent.children, key=attrgetter('sort_order')).index(self) + 1
		return -1

	def check_markup(self, value: str|int|bool) -> None:
		value = str(value)
		self.render_str(value)

	def check_markup_dict(self, check_dict: dict[str, str|int|bool], *keys: str) -> None:
		for key in keys:
			if key in check_dict:
				self.check_markup(check_dict[key])

	def check_properties(self, kwargs: dict[str, str|int|bool]) -> None:
		self.check_markup_dict(kwargs, 'title')

	def update_properties(self, kwargs: dict[str, str|int|bool]) -> None:
		if 'classes' in kwargs:
			self.set_classes(str(kwargs['classes']))
		if 'title' in kwargs:
			self.title = str(kwargs['title']) if kwargs['title'] else self.title_from_id()
		if 'scroll_on_activity' in kwargs:
			scroll_on_activity = kwargs['scroll_on_activity']
			if isinstance(scroll_on_activity, int):
				self.scroll_on_activity = scroll_on_activity
			else:
				self.scroll_on_activity = bool(scroll_on_activity)

	def export_properties(self) -> dict[str, Any]:
		prop: dict[str, Any] = {}
		prop['id'] = self.title_from_id()
		prop['classes'] = ' '.join(self.classes)
		prop['title'] = self.title
		prop['scroll_on_activity'] = self.scroll_on_activity
		return prop

	def save(self, opener: Callable[[str, int], int], filename: str, extra_properties: dict[str, Any]) -> None:
		properties = {**extra_properties, **self.export_properties()}
		with open(filename + '.properties.json', 'w', encoding='utf-8', opener=opener) as properties_filedesc:
			json.dump(properties, properties_filedesc, indent=4)
			properties_filedesc.write('\n')

	def notify_copy_to_clipboard(self, result: bool, explanation: str) -> None:
		step_index = self.index()
		title = f'Step #{step_index} copied!' if result else f'Failed to copy step #{step_index}'
		if explanation:
			message = explanation
		elif result:
			message = f'Step #{step_index} copied to clipboard'
		else:
			message = f'Could not copy step #{step_index} to clipboard'
		self.app.notify(message, title=title, severity='information' if result else 'error')

	@staticmethod
	def copy_to_clipboard(func: Callable) -> Callable:
		def wrapper(self: AbstractStep, *args: Any, **kwargs: Any) -> None:
			try:
				result, data, explanation = func(self, *args, **kwargs)
				if result:
					copy(self.app, data)
			except Exception as exc:
				result, explanation = False, str(exc)
			self.notify_copy_to_clipboard(result, explanation)
		return wrapper

	def activity(self) -> None:
		msg = AbstractStep.StepActivity(self)
		self.call_after_refresh(self.post_message, msg)

	def search_label(self, label: str, search: TextSearch) -> tuple[bool, Text]:
		"""
		Helper to search labels.
		"""
		text = Text.from_markup(label)
		self.label_search_cursor = search.search(text.plain, self.label_search_cursor)
		if self.label_search_cursor is None:
			return False, text
		return True, search.highlight(text, *self.label_search_cursor)

	DEFAULT_COLORS = """
	$step_default: $secondary;
	$step_success: #00ff00; /* bright green */
	$step_warning: #ffa500; /* orange */
	$step_error: #ff6347; /* tomato */
	$step_inactive: #bebebe; /* halfway between lightgray #d3d3d3 and darkgray #a9a9a9 */
	"""
	DEFAULT_CSS = DEFAULT_COLORS + """
	AbstractStep {
		/* If we do not specify "auto" here, each step will take 100% of the viewport: */
		height: auto;
		background: $step_default;
		border-left: blank;
		color: $text;
		&.success { background: $step_success; }
		&.warning { background: $step_warning; }
		&.error { background: $step_error; }
		&.inactive { background: $step_inactive; }
	}
	AbstractStep:focus, AbstractStep:focus-within {
		border-left: thick $primary-lighten-3;
	}
	"""



"""
This file provides a registry that exposes all TUI/Textual classes for Moulti widgets, keyed by their CLI command. This
can be used to determine whether a given command actually makes sense and which class to use to handle it.
All such widgets are expected to provide two modules (cli and tui) along with the following attributes:
- cli.COMMAND: str
- cli.add_cli_arguments(subparsers: argparse._SubParsersAction) -> None
- tui.MoultiWidgetClass: subclass of AbstractStep
"""
from moulti.widgets import WIDGET_MODULES
from .abstractstep.tui import AbstractStep

Registry = dict[str, type[AbstractStep]]

class MoultiWidgets:
	_registry: Registry = {}

	@classmethod
	def register(cls, command: str, widget_class: type[AbstractStep]) -> None:
		cls._registry[command] = widget_class

	@classmethod
	def register_widget_module(cls, module_name: str) -> None:
		cli = __import__(f'moulti.widgets.{module_name}.cli', globals(), locals(), 'COMMAND')
		tui = __import__(f'moulti.widgets.{module_name}.tui', globals(), locals(), 'MoultiWidgetClass')
		cls.register(cli.COMMAND, tui.MoultiWidgetClass)

	@classmethod
	def register_all_widget_modules(cls) -> None:
		for widget_module in WIDGET_MODULES:
			cls.register_widget_module(widget_module)

	@classmethod
	def registry(cls) -> Registry:
		if not cls._registry:
			cls.register_all_widget_modules()
		return cls._registry

	@classmethod
	def command_to_class(cls, command: str) -> type[AbstractStep] | None:
		if not cls._registry:
			cls.register_all_widget_modules()
		return cls._registry.get(command)

	@classmethod
	def class_to_command(cls, widget_class: type[AbstractStep]) -> str | None:
		for command, moulti_widget_class in cls._registry.items():
			if moulti_widget_class == widget_class:
				return command
		return None



"""
Provide everything to deal with ANSI themes in Textual applications.
"""
import re
from os import environ
from types import ModuleType
from rich import terminal_theme
from textual.app import App as TextualApp
from textual.filter import ANSIToTruecolor, LineFilter
from .environ import env

TerminalTheme = terminal_theme.TerminalTheme
NamedTerminalTheme = tuple[str, TerminalTheme]
TerminalThemeCollection = dict[str, TerminalTheme]

COLOR_RE = r'#?(?P<color>[0-9a-f]{6})'
OVERRIDE_RE = r'^(?P<name>(?:dark|light)[bf]g)=' + COLOR_RE + '$'

def dump_filters(app: TextualApp) -> str:
	if not hasattr(app, '_filters'):
		return 'App._filters no longer exists'
	app_filters = getattr(app, '_filters')
	return ', '.join((type(line_filter).__name__ for line_filter in app_filters))

def replace_line_filter(app: TextualApp, filter_type: type[LineFilter], new_filter: LineFilter|None) -> bool:
	if not hasattr(app, '_filters'):
		return False
	app_filters = getattr(app, '_filters')
	for index, textual_filter in enumerate(app_filters):
		if isinstance(textual_filter, filter_type):
			if new_filter is None:
				del app_filters[index]
			else:
				app_filters[index] = new_filter
			break
	else:
		if new_filter is not None:
			app_filters.insert(0, new_filter)
	return True

def ansi_themes_from_module(source: ModuleType) -> TerminalThemeCollection:
	TerminalThemeClass = TerminalTheme
	themes = {}
	for name, obj in source.__dict__.items():
		if isinstance(obj, TerminalThemeClass):
			themes[name] = obj
	return themes

def ansi_themes_from_rich() -> TerminalThemeCollection:
	return ansi_themes_from_module(terminal_theme)

def ansi_themes_from_textual() -> TerminalThemeCollection:
	try:
		textual = __import__('textual._ansi_theme')
		textual_ansi_themes = textual._ansi_theme # pylint: disable=protected-access
		return ansi_themes_from_module(textual_ansi_themes)
	except Exception:
		return {}

def color_from_string(desc: str) -> tuple[int, int, int]:
	"""
	Turn a 6-hexdigit string into three integers.
	"""
	return int(desc[0:2], 16), int(desc[2:4], 16), int(desc[4:6], 16)

def ansi_theme_from_string(description: str) -> TerminalTheme|None:
	"""
	Return a TerminalTheme object based on a complete single-line description.
	Example of an 8-color theme: bg=000000,fg=ffffff,ansi=ffffff:80ffff:ff80ff:8080ff:ffff80:80ff80:ff8080:3f3f3f
	bg and fg define background and foreground colors, respectively.
	ansi must be a colon-separated list of exactly 8 or exactly 16 colors.
	Colors may optionally start with '#'.
	"""
	background, foreground, colors = None, None, []
	for part in description.split(','):
		part = part.lower()
		if rem := re.match(f'^bg={COLOR_RE}$', part):
			background = color_from_string(rem.group('color'))
		elif rem := re.match(f'^fg={COLOR_RE}$', part):
			foreground = color_from_string(rem.group('color'))
		elif part.startswith('ansi='):
			for color_part in part[5:].split(':'):
				if rem := re.match(f'^{COLOR_RE}$', color_part):
					colors.append(color_from_string(rem.group('color')))
	if background is not None and foreground is not None and len(colors) in (8, 16):
		if len(colors) == 8:
			return TerminalTheme(background, foreground, colors)
		return TerminalTheme(background, foreground, colors[:8], colors[8:])
	return None

def ansi_themes_from_environment(environment_prefix: str = '') -> TerminalThemeCollection:
	"""
	Example of supported environment variables (without prefix):
	ANSI_THEME_8COLORS=bg=000000,fg=ffffff,ansi=ffffff:80ffff:ff80ff:8080ff:ffff80:80ff80:ff8080:3f3f3f
	bg and fg define background and foreground colors, respectively.
	ansi must be a colon-separated list of 8 or 16 colors.
	Colors may optionally start with '#'.
	"""
	themes = {}
	prefix = environment_prefix + 'ANSI_THEME_'
	for env_name, env_value in environ.items():
		if env_name.startswith(prefix):
			theme_name = env_name[len(prefix):]
			theme_obj = ansi_theme_from_string(env_value)
			if theme_obj is not None:
				themes[theme_name] = theme_obj
	return themes

def all_ansi_themes(environment_prefix: str = '') -> TerminalThemeCollection:
	themes = ansi_themes_from_textual()
	for name, theme in ansi_themes_from_rich().items():
		themes[name] = theme
	for name, theme in ansi_themes_from_environment(environment_prefix).items():
		themes[name] = theme
	return themes


class AnsiThemePolicy:
	def __init__(self, dark_theme: NamedTerminalTheme|None, light_theme: NamedTerminalTheme|None) -> None:
		self.dark_theme = dark_theme
		self.light_theme = light_theme
		self.overrides: dict[str, tuple[int, int, int]] = {}
		self.verbatim = False

	def apply(self, app: TextualApp) -> bool:
		if self.dark_theme is not None:
			app.ansi_theme_dark = self.dark_theme[1]
		if self.light_theme is not None:
			app.ansi_theme_light = self.light_theme[1]
		for name, color in self.overrides.items():
			theme = app.ansi_theme_dark if name.startswith('dark') else app.ansi_theme_light
			attribute = 'background_color' if name.endswith('bg') else 'foreground_color'
			setattr(theme, attribute, color)
		if self.verbatim:
			# Verbatim: remove Textual's ANSIToTruecolor filter:
			replace_line_filter(app, ANSIToTruecolor, None)
		return True

	def __repr__(self) -> str:
		dark_theme = 'None' if self.dark_theme is None else self.dark_theme[0]
		light_theme = 'None' if self.light_theme is None else self.light_theme[0]
		verbatim = self.verbatim
		overrides = self.overrides
		return f'ANSI theme policy: {dark_theme=}, {light_theme=}, {verbatim=}, {overrides=}'

	@classmethod
	def from_string(cls, themes: TerminalThemeCollection, policy: str) -> 'AnsiThemePolicy':
		"""
		Return an AnsiThemePolicy object based on a list of known themes and a policy string.
		A policy string contains a policy ("theme" or "verbatim") followed by a colon (":") followed by a comma-separated
		list of policy-specific options.
		Options for policy "theme":
		- dark=THEMELIST
		- light=THEMELIST
		where THEMELIST is a colon-separated list of known theme names by decreasing order of preference.
		Example: theme:dark=MY_DARK_THEME:DEFAULT_TERMINAL_THEME,light=NIGHT_OWLISH
		Options for policy "verbatim":
		- darkbg=COLOR
		- darkfg=COLOR
		- lightbg=COLOR
		- lightfg=COLOR
        Example: verbatim:darkbg=000000,darkfg=e6e6e6,lightbg=ffaaff,lightfg=000000
		"verbatim" is also accepted as is.
		"""
		def first_known(priority_list: list[str]) -> NamedTerminalTheme|None:
			for theme_name in priority_list:
				if theme_name in themes:
					return (theme_name, themes[theme_name])
			return None

		if policy.startswith("theme:"):
			policy = policy[6:]
			dark_theme, light_theme = None, None
			for part in policy.split(','):
				if part.startswith('dark='):
					dark_theme = first_known(part[5:].split(':'))
				elif part.startswith('light='):
					light_theme = first_known(part[6:].split(':'))
			return AnsiThemePolicy(dark_theme, light_theme)
		verbatim_policy = AnsiThemePolicy(None, None)
		verbatim_policy.verbatim = True
		if policy.startswith("verbatim:"):
			policy = policy[9:]
			for part in policy.split(','):
				if rem := re.match(OVERRIDE_RE, part):
					verbatim_policy.overrides[rem.group('name')] = color_from_string(rem.group('color'))
		return verbatim_policy

	@classmethod
	def from_environment(cls, environment_prefix: str = '') -> 'AnsiThemePolicy':
		"""
		Return an AnsiThemePolicy object based on (optionally prefixed) environment variables ANSI and ANSI_THEME_*.
		"""
		env_var_name = environment_prefix + 'ANSI'
		env_var_value = env(env_var_name, '')
		themes = all_ansi_themes(environment_prefix) if env_var_value.startswith('theme:') else {}
		return cls.from_string(themes, env_var_value)



"""
This file provides a helper that lets all Moulti widgets declare their CLI arguments.
"""
from argparse import _SubParsersAction
from moulti.widgets import WIDGET_MODULES

def add_cli_arguments(subparsers: _SubParsersAction) -> None:
	for module_name in WIDGET_MODULES:
		cli = __import__(f'moulti.widgets.{module_name}.cli', globals(), locals(), 'add_cli_arguments')
		cli.add_cli_arguments(subparsers)



"""
A simple "askpass" program for Moulti, compatible with ssh, Ansible and sudo.
"""
import re
import sys
import random
from .client import Socket, moulti_connect, send
from .environ import env

class NoAnswerException(Exception):
	pass

def generate_step_id(prefix: str = '') -> str:
	return prefix + str(random.randint(10_000_000, 99_999_999))

def get_default_prompt() -> str:
	# If moulti-askpass was invoked with a suffix, turn it into a prompt:
	if sys.argv and (rem := re.search(r'moulti-askpass-(?P<suffix>[^/]+)$', sys.argv[0])):
		return rem.group('suffix').replace('-', ' ').capitalize() + '?'
	# Otherwise, check environment variables; use a dummy prompt as a last resort:
	return env('MOULTI_ASKPASS_DEFAULT_PROMPT', 'askpass')

def get_prompt() -> tuple[str, str, bool]:
	# Assume the user is prompted for a secret value:
	ask_secret = True
	# Ideally, the prompt should be provided as first command-line argument:
	prompt = sys.argv[1] if len(sys.argv) >= 2 else get_default_prompt()
	# Ansible-specific:
	if prompt == '--vault-id' and len(sys.argv) >= 3:
		vault_id = sys.argv[2]
		prompt = f'Password for vault {vault_id}?'
	else:
		# SSH-specific: askpass is also used to accept or refuse host key fingerprints:
		ask_secret = 'fingerprint' not in prompt
	# Escape '[' characters so Textual does not interpret them as Rich tags:
	prompt = prompt.replace('[', r'\[')
	# For proper alignment, use the first line as title and the rest as top-text:
	split = prompt.split('\n', 1)
	if len(split) == 1:
		split.append('')
	return split[0], split[1], ask_secret

def get_answer(socket: Socket, step_type: str, step_id: str) -> str:
	try:
		send(socket, {'command': 'scroll', 'id': step_id})
	except Exception:
		pass

	try:
		reply = send(socket, {'command': step_type, 'action': 'get-answer', 'id': step_id, 'wait': True})
	except Exception as exc:
		raise NoAnswerException() from exc

	try:
		send(socket, {'command': step_type, 'action': 'delete', 'id': step_id})
	except Exception:
		pass

	if not reply['done'] or 'answer' not in reply:
		raise NoAnswerException()

	return reply['answer']

def main() -> None:
	try:
		prefix = 'askpass_'
		main_input_id = generate_step_id(prefix)
		title, top_text, ask_secret = get_prompt()
		common = {
			'action': 'add', 'id': main_input_id, 'classes': 'askpass', 'title': title, 'top_text': top_text,
			'bottom_text': ' ', 'collapsed': False,
		}
		ssh_askpass_prompt = env('SSH_ASKPASS_PROMPT')
		ok_button = ['ok', 'success', 'OK']
		cancel_button = ['cancel', 'error', 'Cancel']

		with moulti_connect() as moulti_socket:
			if ssh_askpass_prompt == 'none':
				send(moulti_socket, {'command': 'buttonquestion', 'button': [ok_button], **common})
				get_answer(moulti_socket, 'buttonquestion', main_input_id)
			elif ssh_askpass_prompt == 'confirm':
				send(moulti_socket, {'command': 'buttonquestion', 'button': [ok_button, cancel_button], **common})
				answer = get_answer(moulti_socket, 'buttonquestion', main_input_id)
				sys.exit(int(answer == 'cancel')) # ok => exit 0, cancel => exit 1
			else:
				send(moulti_socket, {'command': 'inputquestion', 'password': ask_secret, **common})
				answer = get_answer(moulti_socket, 'inputquestion', main_input_id)
				print(answer)
		sys.exit(0)
	except NoAnswerException:
		sys.exit(1)
	except KeyboardInterrupt:
		print('')
		sys.exit(10)
	except Exception:
		sys.exit(15)

if __name__ == '__main__':
	main()



from argparse import ArgumentParser
from moulti.environ import bool_or_int

def add_abstractstep_options(parser: ArgumentParser, none: bool = False) -> None:
	"""Options common to step add (with actual default values) and step update (with None default values)."""
	parser.add_argument('--title', type=str, default=None, help='step title, always visible')
	parser.add_argument('--classes', '-c', type=str, default=None if none else 'standard', help='step class (color): standard, error, warning, success')
	parser.add_argument('--scroll-on-activity', type=bool_or_int, default=None if none else False, help='whether to scroll to this step upon activity', metavar='{true,false,...,-2,-1,0,1,2,...}')



from .common import moulti_test
assert moulti_test

COMMAND = ['maximize.bash']

def test_maximize_small_step(moulti_test):
	# Maximize the first step:
	assert moulti_test(command=COMMAND, press=('tab', 'tab', 'f'))

def test_unmaximize_small_step(moulti_test):
	# Maximize then unmaximize the first step:
	assert moulti_test(command=COMMAND, press=('tab', 'tab', 'f', 'f'))

def test_maximize_large_step(moulti_test):
	# Maximize the last step:
	assert moulti_test(command=COMMAND, press=('shift+tab', 'f'))

def test_unmaximize_large_step(moulti_test):
	# Maximize then unmaximize the last step:
	assert moulti_test(command=COMMAND, press=('shift+tab', 'f', 'f'))

def test_search_maximize(moulti_test):
	press = (
		'tab', 'tab', 'f', # maximize the first step
		'/', *r'\b5\b', 'enter', # search it for a standalone '5' character
		'f', # unmaximize it
		'n', # ask for the next result, which is expected in the second step
	)
	assert moulti_test(command=COMMAND, press=press)

def test_search_unmaximize(moulti_test):
	press = (
		'/', *r'\b5\b', 'enter', # search for a standalone '5' character, which is expected in the first step
		'tab', 'tab', 'tab', 'f', # maximize the second step
		'n', # search it for the next result
		'f', # unmaximize it
		'n', # ask for the next result, which is expected in the third step
	)
	assert moulti_test(command=COMMAND, press=press)



from moulti.environ import env, enva, envd, pint

def test_env(monkeypatch):
	monkeypatch.delenv('NON_EXISTENT_ENV_VAR', False)
	assert env('NON_EXISTENT_ENV_VAR', 'default') == 'default'

	monkeypatch.setenv('EMPTY', '')
	assert env('EMPTY', 'default') == ''

	monkeypatch.setenv('NAME', 'value')
	assert env('NAME', 'default') == 'value'

	monkeypatch.setenv('INTEGER', '123')
	assert env('INTEGER', -1, [int]) == 123

	monkeypatch.setenv('NON_INTEGER', 'foo')
	assert env('NON_INTEGER', -1, [int]) == -1
	assert env('NON_INTEGER', -1, [int, str]) == 'foo'
	assert env('NON_INTEGER', -1, [str, int]) == 'foo'


def test_enva(monkeypatch):
	monkeypatch.delenv('NON_EXISTENT_ENV_VAR', False)
	assert enva('NON_EXISTENT_ENV_VAR', [1, 2, 3]) == [1, 2, 3]

	monkeypatch.setenv('EMPTY', '')
	assert enva('EMPTY', [1, 2, 3]) == []

	monkeypatch.setenv('INTEGERS', '-2,-1,0,1,2')
	assert enva('INTEGERS', [1, 2, 3]) == ['-2', '-1', '0', '1', '2']

	monkeypatch.setenv('INTEGERS', '-2,-1,0,1,2')
	assert enva('INTEGERS', [1, 2, 3], [int]) == [-2, -1, 0, 1, 2]

	monkeypatch.setenv('INTEGERS', '-2,-1,0,1,2')
	assert enva('INTEGERS', [1, 2, 3], [pint]) == [1, 2, 3]

	monkeypatch.setenv('INTEGERS', '-2,-1,0,1,2')
	assert enva('INTEGERS', [1, 2, 3], [pint], strict=False) == [0, 1, 2]

	monkeypatch.setenv('INTEGERS', '-2,-1,0,1,2')
	assert enva('INTEGERS', [1, 2, 3], [pint, str]) == ['-2', '-1', 0, 1, 2]

	monkeypatch.setenv('INTEGERS', '-2;-1;0;1;2')
	assert enva('INTEGERS', [1, 2, 3], [int], sep=';') == [-2, -1, 0, 1, 2]

	# Separator with multiple characters:
	monkeypatch.setenv('INTEGERS', '-2;;-1;;0;;1;;2')
	assert enva('INTEGERS', [1, 2, 3], [int], sep=';;') == [-2, -1, 0, 1, 2]

	monkeypatch.setenv('INTEGERS', '-2;;-1;;0;;1;2')
	assert enva('INTEGERS', [1, 2, 3], [int], sep=';;') == [1,2,3]

	monkeypatch.setenv('INTEGERS', '-2;;-1;;0;;1;2')
	assert enva('INTEGERS', [1, 2, 3], [int], sep=';;', strict=False) == [-2, -1, 0]

	# Trailing separator:
	monkeypatch.setenv('INTEGERS', '-2;;-1;;0;;1;2;;')
	assert enva('INTEGERS', [1, 2, 3], [int], sep=';;', strict=False) == [-2, -1, 0]

	monkeypatch.setenv('INTEGERS', '-2;;-1;;0;;1;2;;')
	assert enva('INTEGERS', [1, 2, 3], [int], sep=';;', strict=True) == [1, 2, 3]


def test_envd(monkeypatch):
	monkeypatch.delenv('NON_EXISTENT_ENV_VAR', False)
	assert envd('NON_EXISTENT_ENV_VAR', {'a': 'b'}) == {'a': 'b'}

	monkeypatch.setenv('EMPTY', '')
	assert envd('EMPTY', {'a': 'b'}) == {}

	monkeypatch.setenv('INTEGERS', 'a=1,b=2,c=3')
	assert envd('INTEGERS') == {'a': '1', 'b': '2', 'c': '3'}

	monkeypatch.setenv('INTEGERS', 'a=1,b=2,c=3')
	assert envd('INTEGERS', types=[int]) == {'a': 1, 'b': 2, 'c': 3}

	monkeypatch.setenv('INTEGERS', 'a=1,b=2,c=3,d=foo')
	assert envd('INTEGERS', types=[int]) is None

	monkeypatch.setenv('INTEGERS', 'a=1,b=2,c=3,d=foo')
	assert envd('INTEGERS', types=[int], strict=False) == {'a': 1, 'b': 2, 'c': 3}

	monkeypatch.setenv('INTEGERS', 'a=1,b=2,c=3,d=foo=')
	assert envd('INTEGERS', types=[int, str]) == {'a': 1, 'b': 2, 'c': 3, 'd': 'foo='}

	monkeypatch.setenv('INTEGERS', 'a==1,b==2,c==3,d==foo==')
	assert envd('INTEGERS', types=[int, str], sepkv='==') == {'a': 1, 'b': 2, 'c': 3, 'd': 'foo=='}

	monkeypatch.setenv('INTEGERS', 'a=1,b,c=3')
	assert envd('INTEGERS', types=[int], sepkv='==') is None



# List of moulti.widgets.* modules that provide non-abstract widgets that can be shown in Moulti.
WIDGET_MODULES = [
	'step',
	'buttonquestion',
	'inputquestion',
	'question',
	'divider',
]

class MoultiWidgetException(Exception):
	"""
	Generic exception that widgets may raise.
	"""

class MoultiWidgetInvalidPropertyException(MoultiWidgetException):
	"""
	Exception widgets may raise to indicate they were passed an invalid property.
	"""



import os
import sys
import json
import time
import asyncio
import datetime
import selectors
import subprocess
from concurrent.futures import ThreadPoolExecutor
from queue import Queue
from typing import Any, Callable, Iterator, cast
from socket import socket as Socket
from shutil import which
from time import time_ns, localtime, strftime
from threading import get_ident, Lock
from rich.markup import MarkupError, escape
from rich.text import Text
from textual import work
from textual import __version__ as TEXTUAL_VERSION
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.css.query import NoMatches
from textual.dom import BadIdentifier
from textual.reactive import reactive
from textual.widgets import Label, ProgressBar
from textual.worker import get_current_worker, NoActiveWorker
from . import __version__ as MOULTI_VERSION
from .ansi import AnsiThemePolicy, dump_filters
from .environ import env, envd, spint, existing_filepath
from .helpers import call_all, clean_selector
from .search import TextSearch, MatchSpan
from .security import MoultiSecurityPolicy
from .server import Message, FDs, MoultiServer, current_instance
from .themes import MOULTI_THEMES
from .widgets import MoultiWidgetException
from .widgets.tui import MoultiWidgets
from .widgets.footer import Footer
from .widgets.stepcontainer import StepContainer
from .widgets.abstractstep.tui import AbstractStep
from .widgets.collapsiblestep.tui import CollapsibleStep
from .widgets.step.tui import Step
from .widgets.moulticonsole import MoultiConsole
from .widgets.quitdialog import QuitDialog
from .widgets.searchinput import SearchInput, SearchInputWidget
from .widgets.helpscreen import HelpScreen


MOULTI_RUN_OUTPUT_STEP_ID = 'moulti_run_output'

def timestamp() -> str:
	timestamp_ns = time_ns()
	timestamp_s = timestamp_ns//10**9
	ns = timestamp_ns - timestamp_s * 10**9
	ms = ns // 10**6
	timestr = strftime('%FT%T', localtime(timestamp_s))
	timestr = f'{timestr}.{ms:03d} '

	return timestr

def is_ansible(command: list[str]) -> bool:
	return bool(len(command)) and 'ansible' in command[0]

def add_abs_path_to_environment(environment: dict[str, str], env_var: str, command: str) -> bool:
	if abs_path := which(command):
		environment[env_var] = abs_path
		return True
	return False

def run_environment(command: list[str], socket_path: str, copy: bool = True) -> dict[str, str]:
	"""
	Return environment variables set by "Moulti run <command>".
	"""
	environment = os.environ.copy() if copy else {}

	environment['MOULTI_RUN'] = 'moulti'
	environment['MOULTI_SOCKET_PATH'] = socket_path
	environment['MOULTI_INSTANCE_PID'] = str(os.getpid())

	if env('SSH_ASKPASS') is None:
		environment['SSH_ASKPASS'] = 'moulti-askpass'
		environment['SSH_ASKPASS_REQUIRE'] = 'force'

	if env('SUDO_ASKPASS') is None:
		# sudo requires an absolute path:
		add_abs_path_to_environment(environment, 'SUDO_ASKPASS', 'moulti-askpass')

	if (ansible_policy := env('MOULTI_ANSIBLE', 'default')) != 'no':
		if ansible_policy == 'force' or (is_ansible(command) and env('ANSIBLE_STDOUT_CALLBACK') is None):
			from . import ansible # pylint: disable=import-outside-toplevel
			if ansible_callback_paths := env('ANSIBLE_CALLBACK_PLUGINS', ''):
				ansible_callback_paths += ':'
			ansible_callback_paths += ansible.__path__[0]
			environment['ANSIBLE_CALLBACK_PLUGINS'] = ansible_callback_paths
			environment['ANSIBLE_STDOUT_CALLBACK'] = 'moulti'
			environment['ANSIBLE_FORCE_COLOR'] = 'yes'
			# ansible-playbook may still read from /dev/tty through getpass.getpass():
			ansible_askpass = {
				'ANSIBLE_CONNECTION_PASSWORD_FILE': {'args': ('-k', '--ask-pass'), 'suffix': 'connection-password'},
				'ANSIBLE_BECOME_PASSWORD_FILE': {'args': ('-K', '--ask-become-pass'), 'suffix': 'become-password'},
				'ANSIBLE_VAULT_PASSWORD_FILE': {
					'args': ('-J', '--ask-vault-pass', '--ask-vault-password'),
					'suffix': 'vault-client',
				},
			}
			for env_var, conf in ansible_askpass.items():
				if env_var in environment:
					continue
				for arg in conf['args']:
					if arg in command:
						add_abs_path_to_environment(environment, env_var, f'moulti-askpass-{conf["suffix"]}')
						del command[command.index(arg)]

	return environment

class MoultiMessageException(Exception):
	pass

class Moulti(App):
	"""
	moulti is a Terminal User Interface (TUI) meant to display multiple outputs.
	It is visually similar to https://github.com/dankilman/multiplex but unlike
	it, moulti does not do anything until instructed to.
	"""

	# The header and search input widgets should remain visible in maximized mode:
	ALLOW_IN_MAXIMIZED_VIEW = '#header,SearchInputWidget,' + App.ALLOW_IN_MAXIMIZED_VIEW

	BINDINGS = [
		("s", "save", "Save"),
		Binding("z", "toggle_console", "Console"),
		("x", "collapse_all(False)", "Expand all"),
		("o", "collapse_all(True)", "Collapse all"),
		Binding("X", "collapse_new(False)", "Expand new widgets", show=False),
		Binding("O", "collapse_new(True)", "Collapse new widgets", show=False),
		Binding("ctrl+x", "collapse_both(False)", "Expand", show=False),
		Binding("ctrl+o", "collapse_both(True)", "Collapse", show=False),
		Binding("d", "light_mode", "Light"),
		Binding("d", "dark_mode", "Dark"),
		("h", "help", "Help"),
		("q", "quit", "Quit"),
		Binding("/", "search(True)", "Search forward", show=False),
		Binding("?", "search(False)", "Search backward", show=False),
		Binding("N", "search_previous(False)", "Previous search result", show=False),
		Binding("n", "search_next(False)", "Next search result", show=False),
	]

	# Disable Textual's command palette; it may come back if Moulti ends up with too many commands though:
	ENABLE_COMMAND_PALETTE = False
	# Do not minimize on 'escape' because widgets need to restore their max height:
	ESCAPE_TO_MINIMIZE = False

	dark = reactive(True)
	"""True for dark mode, False for light mode."""

	def __init__(self, command: list[str]|None = None, instance_name: str|None = None):
		self.instance_name = instance_name or current_instance()
		self.server: MoultiServer|None = None
		self.security = MoultiSecurityPolicy(self)
		self.init_quit_policy()
		self.init_widgets()
		self.end_user_console = MoultiConsole('moulti_console', classes='hidden')
		self.init_command = command
		self.init_command_running: bool|None = None
		"""Whether the init command is running: None until it starts, True while it runs, False afterwards."""
		self.exit_first_policy = ''
		"""What to do when "moulti run" must exit before the process it launched? 'terminate' terminates the process,
		any other value leaves it running in the background."""
		self.save_lock = Lock()
		"""
		Lock that ensures Moulti does not generate multiple exports of the current instance at a time.
		"""
		self.enforce_collapsible: bool|None = self.init_enforce_collapsible()
		"""
		Whether to enforce the state of new collapsible widgets.
		None: do not enforce
		True: force-collapse new collapsible widgets
		False: force-expand new collapsible widgets
		"""
		self.current_search: TextSearch|None = None
		"""
		Current text search.
		"""
		self.title_search_cursor: MatchSpan = None
		self.search_cursor: bool|None = None
		"""
		False: search the title; True: search steps; None: decide based on search.next_result.
		"""
		super().__init__()
		self.dark = env('MOULTI_MODE', 'dark') != 'light'
		self.set_title('Moulti')

	def init_quit_policy(self) -> None:
		"""
		Initialize the quit policy, i.e. how Moulti should behave when end users request to quit.
		"""
		# Default quit policy: ask if the driving process is still running, otherwise quit directly:
		self.quit_policy = {'running': 'ask', 'default': 'quit'}
		# The default policy can be overridden through the environment variable MOULTI_QUIT_POLICY:
		env_overrides = envd('MOULTI_QUIT_POLICY', {}, sepa=';')
		if (override := env_overrides.get('running')) in ('ask', 'leave', 'terminate'):
			self.quit_policy['running'] = override
		if (override := env_overrides.get('default')) in ('ask', 'quit'):
			self.quit_policy['default'] = override

	def init_widgets(self) -> None:
		self.title_label = Label('Title', id='header')
		self.title_label.tooltip = f'Instance name: {self.instance_name}'
		self.steps_container = StepContainer()
		self.search_input = SearchInputWidget(id='search_bar')
		self.progress_bar = ProgressBar(id='progress_bar', show_eta=False)
		self.footer = Footer()

	def init_enforce_collapsible(self) -> bool|None:
		mec_env_var = env('MOULTI_ENFORCE_COLLAPSIBLE')
		if mec_env_var == 'collapse':
			return True
		if mec_env_var == 'expand':
			return False
		return None

	def set_title(self, title: str) -> None:
		self.title = title
		self.title_label.update(title)

	def setup_ansi_behavior(self) -> None:
		"""
		By default and by design, Textual strives to avoid outputting ANSI colors (i.e. the 16 colors that have no
		official value). This makes perfect sense for most applications. Moulti is special because its users expect
		steps to display the exact same thing as their terminal with no or little care for portability.
		"""
		if env('MOULTI_ANSI') == 'textual_default':
			# MOULTI_ANSI=textual_default means Moulti neither loads, applies nor alters any theme.
			# However, Textual default themes will apply, unaltered.
			return
		policy = AnsiThemePolicy.from_environment('MOULTI_')
		self.logconsole(f'{policy}')
		policy.apply(self)
		self.logconsole(f'Textual filters: {dump_filters(self) or "none"}')

	def init_threads(self) -> None:
		max_pass = env('MOULTI_PASS_CONCURRENCY', default=20, types=(spint,))
		default_threads = 2 # network_loop() + exec()
		threads_per_pass = 2 # Each "moulti pass" command spawns 2 threads -- cf Step.cli_action_pass()
		thread_count = default_threads + threads_per_pass * max_pass
		moulti_executor = ThreadPoolExecutor(max_workers=thread_count, thread_name_prefix='moulti')
		asyncio.get_running_loop().set_default_executor(moulti_executor)
		self.logconsole(f'allow up to {thread_count} threads for up to {max_pass} concurrent "moulti pass" commands')

	def compose(self) -> ComposeResult:
		"""Create child widgets for the app."""
		yield self.title_label
		yield self.steps_container
		yield self.search_input
		yield self.progress_bar
		yield self.footer
		yield self.end_user_console

	def on_mount(self) -> None:
		for theme in MOULTI_THEMES.values():
			self.register_theme(theme)
		self.theme = 'moulti-dark' if self.dark else 'moulti-light'

	def on_ready(self) -> None:
		self.logconsole(f'Moulti v{MOULTI_VERSION}, Textual v{TEXTUAL_VERSION}, Python v{sys.version}')
		self.logconsole(f'instance "{self.instance_name}", PID {os.getpid()}')
		self.logconsole('quit policy: ' + ', '.join(f"{k}='{v}'" for k,v in self.quit_policy.items()))
		self.setup_ansi_behavior()
		self.init_threads()
		widget_list = ' '.join(MoultiWidgets.registry().keys())
		self.logconsole(f'known widgets: {widget_list}')
		self.network_loop()

	def on_search_input_cancel_search(self, _event: SearchInput.CancelSearch) -> None:
		if self.screen.maximized:
			self.screen.maximized.focus()

	def on_search_input_widget_new_search(self, message: SearchInputWidget.NewSearch) -> None:
		self.current_search = message.search
		if self.screen.maximized:
			self.screen.maximized.focus()
		self.search(self.current_search)

	def search_title(self, search: TextSearch) -> bool:
		text = Text.from_markup(self.title)
		self.title_search_cursor = search.search(text.plain, self.title_search_cursor)
		if found := self.title_search_cursor is not None:
			search.highlight(text, *self.title_search_cursor)
		self.title_label.update(text)
		return found

	@work(exclusive=True, group='search', name='search')
	async def search(self, search: TextSearch) -> bool:
		if self.search_cursor is None:
			self.search_cursor = not search.next_result
			parts_to_search = 2 # Only 2 parts to search here
		else:
			# This is a 3-part search:
			# 1. from the latest result to the end of the current part
			# 2. inside the other part
			# 3. from the start of the current part to the latest result
			parts_to_search = 3
		for _ in range(parts_to_search):
			found = self.steps_container.search(search) if self.search_cursor else self.search_title(search)
			if found:
				return True
			self.search_cursor = not self.search_cursor
		search_str = escape(str(search))
		self.notify(f'{search_str}: not found', title='Text search: pattern not found', severity='error')
		return False

	def network_loop_is_ready(self) -> None:
		if self.init_command is not None:
			self.exec(self.init_command)

	def output_policy(self) -> bool|None:
		"""
		`moulti run` launches an arbitrary command. In an ideal world, this command should not output anything on
		stdout/stderr. In practice, this method returns:
		- None if Moulti should ignore stdout/stderr (default)
		- False if Moulti should discard stdout/stderr
		- True if Moulti should harvest stdout/stderr and append it to the "moulti_run_output" step
		"""
		policy = None
		if value := env('MOULTI_RUN_OUTPUT'):
			if value == 'discard':
				policy = False
			elif value == 'harvest':
				policy = True
		return policy

	def output_policy_popen_args(self) -> dict[str, Any]:
		"""
		Return the adequate subprocess.Popen() arguments for `moulti run` based on output_policy().
		"""
		policy = self.output_policy()
		if policy is None: # default/ignore
			return {'stdout': None, 'stderr': None}
		if policy: # harvest
			return {
				'stdout': subprocess.PIPE,
				'stderr': subprocess.STDOUT,
				'text': False,
			}
		# discard:
		return {'stdout': subprocess.DEVNULL, 'stderr': subprocess.DEVNULL}

	def output_policy_step(self) -> Step:
		"""
		Return the step that should display unexpected stdout/stderr lines harvested by `moulti run`.
		"""
		step = self.step_from_message({'id': MOULTI_RUN_OUTPUT_STEP_ID})
		if step is not None:
			return cast(Step, step)
		# This should not fail since the id is valid and reserved:
		step = Step(id=MOULTI_RUN_OUTPUT_STEP_ID, collapsed=False)
		call = (self.steps_container.add_step, step)
		self.call_from_thread(call_all, [call])
		return step

	def output_policy_pass(self, step: Step, filedesc: Any, initial_data: str) -> None:
		"""
		Set up the queue and threads necessary to display unexpected stdout/stderr lines harvested by `moulti run`.
		"""
		# This method is very similar to Step.cli_action_pass().
		queue: Queue = Queue()
		# Helper functions:
		def reply(**_kwargs: Any) -> None:
			pass
		def debug(line: str) -> None:
			self.logconsole(f'exec: {line}')
		helpers = {'file_descriptors': [filedesc], 'debug': debug, 'reply': reply}
		step.append_from_queue(queue, helpers)
		queue.put_nowait(initial_data)
		step.append_from_file_descriptor_to_queue(queue, {}, helpers)

	@work(thread=True, group='app-exec', name='moulti-run')
	def exec(self, command: list[str]) -> None:
		"""
		Launch the given command with the assumption it is meant to drive the current Moulti instance.
		This method is the heart of `moulti run`.
		"""
		worker = get_current_worker()
		harvest_output = self.output_policy() # Harvest stdout/stderr lines
		output_delegated = False
		output_selector = None
		process = None
		try:
			self.init_command_running = True
			original_command = command.copy()
			self.logconsole(f'exec: about to run {original_command}')
			assert self.server is not None
			environment = run_environment(command, self.server.socket_path)
			popen_args: dict[str, Any] = {'env': environment, 'stdin': subprocess.DEVNULL}
			# Important: run_environment() may have modified command:
			if command != original_command:
				self.logconsole(f'exec: command changed to {command}')
			# Not using 'with' because that waits for the process to exit; pylint: disable=consider-using-with
			process = subprocess.Popen(command, **popen_args, **self.output_policy_popen_args())
			self.logconsole(f'exec: {command} launched with PID {process.pid}')
			returncode = None
			if harvest_output:
				output_selector = selectors.DefaultSelector()
				assert process.stdout is not None # for mypy
				output_selector.register(process.stdout, selectors.EVENT_READ)

			while not worker.is_cancelled:
				if output_selector is not None and output_selector.select(0):
					# If the child process outputs anything other than end-of-file, pass it to a special step:
					assert process.stdout is not None and hasattr(process.stdout, 'raw') # for mypy
					if initial_data := process.stdout.raw.read(1):
						step = self.output_policy_step()
						self.logconsole(f'exec: output detected on fd {process.stdout.fileno()}, harvesting')
						self.output_policy_pass(step, process.stdout, initial_data)
						output_delegated = True
					# We no longer need to watch output in this loop:
					clean_selector(output_selector, close_fds=False, close=True)
					output_selector = None
				returncode = process.poll() # non-blocking wait(), e.g. wait4(process.pid, result_addr, WNOHANG, NULL)
				if returncode is not None:
					self.init_command_running = False
					self.logconsole(f'exec: {command} exited with return code {returncode}')
					break
				time.sleep(0.5)
			if returncode is None:
				self.logconsole(f'exec: {command} is still running (PID {process.pid}) but Moulti is about to exit')
				if self.exit_first_policy == 'terminate':
					process.terminate()
				# else just leave the process running in the background and exit.
		except Exception as exc:
			self.logconsole(f'exec: error running {command}: {exc}')
			self.init_command_running = False
		finally:
			clean_selector(output_selector, close_fds=False, close=True)
			if harvest_output and not output_delegated and process is not None and process.stdout is not None:
				self.logconsole(f'exec: no output harvested, closing child process stdout fd {process.stdout.fileno()}')
				process.stdout.close()

	def all_steps(self) -> Iterator[AbstractStep]:
		return self.steps_container.ordered_steps()

	def export_properties(self) -> dict[str, Any]:
		prop: dict[str, Any] = {}
		prop['title'] = self.title
		prop['progress_bar'] = self.progress_bar.styles.display == 'block'
		prop['progress_target'] = self.progress_bar.total
		prop['progress'] = self.progress_bar.progress
		prop['step_position'] = 'bottom' if self.steps_container.has_class('bottom') else 'top'
		prop['step_direction'] = 'down' if self.steps_container.layout_direction_is_down else 'up'
		return prop

	def logconsole(self, line: str) -> None:
		line = timestamp() + line + '\n'
		if self._thread_id == get_ident():
			self.end_user_console.write(line)
		else:
			try:
				worker = get_current_worker()
			except NoActiveWorker:
				worker = None
			if worker and not worker.is_cancelled:
				self.call_from_thread(self.end_user_console.write, line)

	def action_light_mode(self) -> None:
		self.theme = 'moulti-light'
		self.dark = False
		self.refresh_bindings()

	def action_dark_mode(self) -> None:
		self.theme = 'moulti-dark'
		self.dark = True
		self.refresh_bindings()

	def check_action(self, action: str, _parameters: tuple[object, ...]) -> bool:
		if action == 'light_mode':
			return self.dark
		if action == 'dark_mode':
			return not self.dark
		if self.screen.maximized and action in ('save', 'toggle_console', 'collapse_all'):
			return False
		return True

	def action_toggle_console(self) -> None:
		"""Toggle the console."""
		self.end_user_console.toggle_class('hidden')
		if self.end_user_console.has_class('hidden'):
			# Not ideal but reasonable:
			self.steps_container.focus(False)
			# Work around a rendering issue after the console is hidden:
			# Certainly https://github.com/Textualize/textual/issues/5056
			self.refresh(layout=True)
		else:
			# It is often necessary to scroll the console, so give it the focus:
			self.end_user_console.log_widget.focus(False)

	def action_search(self, next_result: bool) -> None:
		self.search_input.action_pop(next_result)
		# Work around a rendering issue when the search bar appears:
		# Likely https://github.com/Textualize/textual/issues/5056
		self.refresh(layout=True)

	def action_search_previous(self) -> None:
		if self.current_search is not None:
			search_copy = self.current_search.copy()
			search_copy.next_result = not search_copy.next_result
			self.search(search_copy)

	def action_search_next(self) -> None:
		if self.current_search is not None:
			self.search(self.current_search)

	def action_collapse_all(self, collapsed: bool = True) -> None:
		"""Collapse all existing steps."""
		for step in self.steps_container.query('CollapsibleStep').results(CollapsibleStep):
			step.collapsible.collapsed = collapsed

	def action_collapse_new(self, enforce: bool) -> None:
		"""Collapse new steps."""
		action = {True: 'collapsing', False: 'expanding'}
		desc = ''
		if self.enforce_collapsible == enforce:
			desc = f'No longer {action[self.enforce_collapsible]}'
			self.enforce_collapsible = None
		else:
			self.enforce_collapsible = enforce
			desc = action[self.enforce_collapsible].capitalize()
		self.notify(f'{desc} new steps')

	def action_collapse_both(self, collapsed: bool) -> None:
		"""Collapse both existing and new steps."""
		if self.enforce_collapsible != collapsed:
			self.action_collapse_new(collapsed)
		self.action_collapse_all(collapsed)

	def action_help(self) -> None:
		"""Display the help screen."""
		self.push_screen(HelpScreen())

	async def action_quit(self) -> None:
		"""Quit Moulti."""
		mode = 'default'
		message = ''
		allow_terminate = False
		if self.init_command is not None and self.init_command_running:
			mode = 'running'
			message = 'The command passed to "moulti run" is still running.'
			allow_terminate = True
		policy = self.quit_policy[mode]
		if policy == 'ask':
			try:
				self.query_one('QuitDialog', QuitDialog).new_quit_request()
			except NoMatches:
				self.push_screen(QuitDialog(message, allow_terminate))
			return
		# leave/quit, terminate:
		self.exit_first_policy = policy
		await super().action_quit()

	def on_quit_dialog_exit_request(self, exit_request: QuitDialog.ExitRequest) -> None:
		"""Exit Moulti, as instructed by the Quit Dialog."""
		self.exit_first_policy = exit_request.exit_first_policy
		self.exit()

	@work(thread=True, group='app-save', name='save')
	def action_save(self) -> None:
		"""
		Export everything currently shown by the instance as a bunch of files in a directory.
		"""
		if not self.save_lock.acquire(blocking=False): # pylint: disable=consider-using-with
			self.logconsole('save: this instance is already being saved.')
			summary = 'This instance is already being saved.'
			self.notify(summary, title='One at a time!', severity='warning')
			return
		export_dir_fd = None
		try:
			# Determine where to create the export directory:
			basepath = env('MOULTI_SAVE_PATH') or '.'
			# Determine what to name the export directory:
			iso_8601_date = datetime.datetime.now().isoformat().replace(":", "-")
			export_dirname = f'{current_instance()}-{iso_8601_date}'
			export_dirpath = os.path.join(basepath, export_dirname)
			# Create the export directory:
			os.makedirs(export_dirpath)
			# Create an "opener" function that opens files relatively to the export directory:
			export_dir_fd = os.open(export_dirpath, 0)
			def opener(path: str, flags: int) -> int:
				return os.open(path, flags, mode=0o666, dir_fd=export_dir_fd)
			# Proceed with the export itself:
			saved_steps = self.save_steps(opener)
			self.save_properties(opener, ('0'*len(str(saved_steps))) + '-')
			# Notify users:
			summary = f'{saved_steps} steps successfully saved\nDir: {basepath}\nSubdir: {export_dirname}'
			self.notify(summary, title='Saved!')
		except Exception as exc:
			self.logconsole(f'save: failed: {exc}')
			summary = 'An error occurred during saving.\nRefer to the console for more details.'
			self.notify(summary, title='Save failed!', severity='error')
		finally:
			self.save_lock.release()
			if export_dir_fd is not None:
				os.close(export_dir_fd)

	def save_properties(self, opener: Callable[[str,int], int], filename_prefix: str = '') -> None:
		filename = filename_prefix + 'instance.properties.json'
		extra_properties = {'command': 'set'}
		with open(filename, 'w', encoding='utf-8', errors='surrogateescape', opener=opener) as instance_file_desc:
			json.dump({**extra_properties, **self.export_properties()}, instance_file_desc, indent=4)
			instance_file_desc.write('\n')

	def save_steps(self, opener: Callable[[str,int], int]) -> int:
		# Fetch a list of all steps to export them:
		steps = list(self.all_steps())
		# Filenames are numbered from 0 to n; how many characters do we need to write n?
		number_prefix_length = len(str(len(steps)))
		step_index = 0
		for step_index, step in enumerate(steps, start=1):
			command = MoultiWidgets.class_to_command(type(step))
			if command is None:
				continue
			extra_properties = {'command': command, 'action': 'add'}
			filename = f'{step_index:0{number_prefix_length}d}-{step.title_from_id()}'
			step.save(opener, filename, extra_properties)
		return step_index

	def reply(self, connection: Socket, raddr: str, message: dict[str, Any], **kwargs: Any) -> None:
		if self.server is None:
			return
		try:
			# If the original message bears a message id, use it to send a
			# smaller reply. Otherwise, send the original message back with
			# extra data.
			if msgid := message.get('msgid'):
				message = {'msgid': msgid, **kwargs}
			else:
				message = {**message, **kwargs}
			self.logconsole(f'{raddr}: <= message={message}')
			self.server.reply(connection, message)
		except Exception as exc:
			self.logconsole(f'{raddr}: reply: kwargs={kwargs}: {exc}')

	def step_from_message(self, message: Message) -> None | AbstractStep:
		try:
			return self.steps_container.query_one('#step_' + str(message['id']), AbstractStep)
		except Exception:
			return None

	def handle_message(self, connection: Socket, raddr: str, message: Message, file_descriptors: FDs) -> None:
		command = str(message.get('command'))
		# "moulti pass" is a special case that actually means "moulti step pass":
		if command == 'pass':
			command = 'step'
			message['command'], message['action'] = command, message['command']
		# Analyse the message to determine what to call or what error to return:
		finally_reply = True
		calls: list[Any] = []
		error = None
		try:
			command_class = MoultiWidgets.command_to_class(command)
			if command_class is not None:
				step = self.step_from_message(message)
				action = str(message.get('action'))
				# add/update/delete actions are considered common/standard and are thus handled here:
				if action == 'add':
					if step:
						raise MoultiMessageException(f'id {message.get("id")} already in use')
					if 'id' not in message:
						raise MoultiMessageException('missing id')
					# Users may create the 'moulti_run_output' step but it MUST be a step, not a question or a divider:
					if message['id'] == MOULTI_RUN_OUTPUT_STEP_ID and message['command'] != 'step':
						err = f'{MOULTI_RUN_OUTPUT_STEP_ID} is a reserved id that must be assigned to a step'
						raise MoultiMessageException(err)
					if self.enforce_collapsible is not None:
						message['collapsed'] = self.enforce_collapsible
					calls.append((self.steps_container.add_step, command_class(**message)))
				else:
					# All other actions require an existing step:
					if not step:
						raise MoultiMessageException(f'unknown id: {message.get("id")}')
					if not isinstance(step, command_class):
						raise MoultiMessageException(f'{message.get("id")} is not a {command}')
					if action == 'update':
						step.check_properties(message)
						if self.enforce_collapsible is not None and message.get('collapsed') is not None:
							del message['collapsed']
						calls.append((step.update_properties, message))
					elif action == 'delete':
						if step.prevent_deletion:
							err = 'cannot proceed with deletion as '
							err += f'{step.prevent_deletion} ongoing operations prevent it'
							raise MoultiMessageException(err)
						calls.append((self.steps_container.remove_step, step))
					else:
						# For actions other than add/update/delete, the widget class is expected to provide a
						# "cli_action_{action}" method:
						method_name = f'cli_action_{action.replace("-", "_")}'
						method = getattr(step, method_name, False)
						if not callable(method):
							raise MoultiMessageException('unknown action {action}')
						# Provide the method with helper functions that abstract away app/network-specific stuff:
						def reply(**kwargs: Any) -> None:
							self.reply(connection, raddr, message, **kwargs)
						def debug(line: str) -> None:
							self.logconsole(f'{raddr}: {line}')
						helpers = {'reply': reply, 'debug': debug, 'file_descriptors': file_descriptors}
						# The method is expected to return a tuple (callable + arguments):
						call = method(message, helpers)
						# An empty tuple means the method chose to handle everything, reply included:
						if call:
							calls.append(call)
						else:
							finally_reply = False
			elif command == 'set':
				if message.get('title') is not None:
					calls.append((self.set_title, str(message['title'])))
				if message.get('step_position') is not None:
					calls.append((self.steps_container.set_class, message['step_position'] == 'bottom', 'bottom'))
				if message.get('step_direction') is not None:
					is_down = bool(message['step_direction'] != 'up')
					calls.append((setattr, self.steps_container, 'layout_direction_is_down', is_down))
				if message.get('progress_bar') is not None:
					display_progress_bar = 'block' if bool(message['progress_bar']) else 'none'
					calls.append((setattr, self.progress_bar.styles, 'display', display_progress_bar))
				if message.get('progress_target') is not None:
					progress_target = float(message['progress_target'])
					final_progress_target: float|None = None if progress_target <= 0 else progress_target
					calls.append((setattr, self.progress_bar, 'total', final_progress_target))
				if message.get('progress') is not None:
					progress_float = float(message['progress'])
					progress_str = str(message['progress'])
					if progress_str.startswith('+') or progress_str.startswith('-'):
						calls.append((self.progress_bar.advance, progress_float))
					else:
						calls.append((setattr, self.progress_bar, 'progress', progress_float))
			elif command == 'scroll':
				step = self.step_from_message(message)
				if step is None:
					raise MoultiMessageException(f'unknown id: {message.get("id")}')
				try:
					offset = int(message['offset'])
				except Exception:
					offset = True
				calls.append((self.steps_container.scroll_to_step, step, offset))
			elif command == 'ping':
				pass
			else:
				raise MoultiMessageException(f'unknown command {command}')
			# At this stage, the analysis is complete; perform the required action and reply accordingly:
			if calls:
				self.call_from_thread(call_all, calls)
		except (BadIdentifier, MarkupError, MoultiMessageException, MoultiWidgetException, ValueError) as exc:
			# If we catch an exception, then we systematically assume we should handle the reply:
			finally_reply = True
			error = str(exc)
		finally:
			if finally_reply:
				self.reply(connection, raddr, message, done=error is None, error=error)

	@work(thread=True, group='app-network', name='network-loop')
	async def network_loop(self) -> None:
		current_worker = get_current_worker()
		self.server = MoultiServer(
			instance_name=self.instance_name,
			socket_path=None,
			loop_callback=lambda: not current_worker.is_cancelled,
			log_callback=self.logconsole,
			ready_callback=self.network_loop_is_ready,
			security_callback=self.security.check,
			message_callback=self.handle_message,
		)
		try:
			self.server.listen()
		except Exception as exc:
			# A Moulti instance is useless if it cannot listen:
			tip = 'Tip: try setting the MOULTI_INSTANCE environment variable, e.g. MOULTI_INSTANCE=$$ moulti init'
			self.exit(return_code=100, message=f'Fatal: {exc}\n{tip}')
			return
		self.server.network_loop()


	DEFAULT_CSS = """
	/* Styles inherited by all widgets: */
	$scrollbar_background: #b2b2b2;
	$scrollbar_inactive_bar: #686868;
	$scrollbar_active_bar: $primary-darken-1;
	Widget {
		/* By default, in dark mode, scrollbars are not rendered clearly; enforce their colours: */
		scrollbar-background: $scrollbar_background;
		scrollbar-background-hover: $scrollbar_background;
		scrollbar-background-active: $scrollbar_background;
		scrollbar-color: $scrollbar_inactive_bar;
		scrollbar-color-active: $scrollbar_active_bar;
		scrollbar-color-hover: $scrollbar_active_bar;
	}
	#textual-tooltip { /* overrides "Tooltip" since it is more specific */
		background: $surface;
	}

	/* One-line title at the top of the screen: */
	#header {
		text-align: center;
		/* For the title to appear centered, the widget must occupy all available width: */
		width: 100%;
		background: $primary;
		color: $text;
		dock: top; /* relevant in maximized mode */
	}

	/* Show the progress bar as a full-width extension of the footer: */
	#progress_bar {
		display: none; /* Do not display by default */
		background: $primary;
		padding-left: 1;
		padding-right: 1;
		width: 100%;
		&> Bar {
			width: 1fr;
			&> .bar--bar {
				color: $warning;
			}
		}
	}
	"""

	# Allow end users to redefine Moulti's look and feel through an optional
	# custom CSS set through an environment variable:
	CSS_PATH = env('MOULTI_CUSTOM_CSS', types=(existing_filepath,))

def main(command: list[str]|None = None) -> None:
	moulti_app = Moulti(command=command)
	moulti_app.run()
	sys.exit(moulti_app.return_code)

if __name__ == '__main__':
	main()
else:
	app = Moulti()



"""
Moulti themes.
Textual < 0.86.0 used to offer two themes: textual-dark and textual-light,
along with an App.dark reactive attribute to switch between them.
Textual 0.86.0 removed App.dark, introduced more themes, and revamped
semantics, styles and colors.
That breaking change severely affected Moulti. The themes below aim to restore
and preserve Moulti's look'n feel.
"""
from textual.theme import Theme

MOULTI_THEMES = {
	'dark': Theme(
		name="moulti-dark",
		primary="#0178D4",
		secondary="#004578",
		warning="#ffa62b",
		error="#ba3c5b",
		success="#4ebf71",
		accent="#ffa62b",
		foreground="#e0e0e0", #
		background="#1e1e1e", #
		surface="#121212",    #
		panel="#24292f",      #
		boost=None,
		dark=True,
		luminosity_spread = 0.15,
		text_alpha = 0.95,
		variables={
			"footer-key-foreground": "#ffffff",
		},
	),
	'light': Theme(
		name="moulti-light",
		primary="#0178D4",
		secondary="#004578",
		warning="#ffa62b",
		error="#ba3c5b",
		success="#4ebf71",
		accent="#ffa62b",
		foreground="#0a0a0a", #
		background="#f5f5f5", #
		surface="#efefef",    #
		panel="#dce3e8",      #
		boost=None,
		dark=False,
		luminosity_spread = 0.15,
		text_alpha = 0.95,
		variables={
			"footer-key-foreground": "#ffffff",
		},
	),
}



from _pytest.fixtures import FixtureRequest
from os import getpid, environ
from asyncio import create_subprocess_exec as async_exec
from asyncio.subprocess import PIPE
import pytest
from moulti.app import Moulti
from .common import moulti_test
assert moulti_test

PASSWORD = 'The secret!'
INPUT_SECRET = ['ctrl+y', *PASSWORD]
INPUT_YES = ['ctrl+y', *'yes']
HIT_1ST_BUTTON = ['ctrl+y']
HIT_2ND_BUTTON = ['ctrl+y', 'tab']
FINGERPRINT_PROMPT = """The authenticity of host '[example.org]:22 ([12.34.56.78]:22)' can't be established.
ED25519 key fingerprint is SHA256:s2DYbBHs784Xe7/LvXCybJe6ncu+x2Dn9mxLwXQ+TfM.
Are you sure you want to continue connecting (yes/no/[fingerprint])?"""

TESTS = [
	{'cmd': ['moulti-askpass']},
	{'cmd': ['moulti-askpass'], 'env': {'MOULTI_ASKPASS_DEFAULT_PROMPT': 'Env-based custom prompt?'}},
	{'cmd': ['moulti-askpass-become-password']},
	{'cmd': ['moulti-askpass-connection-password']},
	{'cmd': ['moulti-askpass-vault-client']},
	{'cmd': ['moulti-askpass', 'Command-line custom prompt?']},
	{'cmd': ['moulti-askpass', '--vault-id', 'SHINY_VAULT']},
	{'cmd': ['moulti-askpass', FINGERPRINT_PROMPT], 'input': INPUT_YES, 'output': 'yes'},
	{'cmd': ['moulti-askpass'], 'env': {'SSH_ASKPASS_PROMPT': 'none'}, 'input': HIT_1ST_BUTTON, 'output': ''},
	{'cmd': ['moulti-askpass'], 'env': {'SSH_ASKPASS_PROMPT': 'confirm'}, 'input': HIT_1ST_BUTTON, 'output': ''},
	{'cmd': ['moulti-askpass'], 'env': {'SSH_ASKPASS_PROMPT': 'confirm'}, 'input': HIT_2ND_BUTTON, 'output': '', 'rc': 1},
]
for index, test in enumerate(TESTS):
	test['index'] = index

@pytest.mark.parametrize('test', TESTS)
def test_askpass_visual(test, moulti_test, monkeypatch):
	"""Visual tests: ensure the askpass step looks like expected."""
	for name, value in test.get('env', {}).items():
		monkeypatch.setenv(name, value)
	async def run_before(pilot):
		await pilot.pause(.5)
		await pilot.press(*test.get('input', INPUT_SECRET))
	assert moulti_test(command=test['cmd'], run_before=run_before)

@pytest.mark.parametrize('test', TESTS)
async def test_askpass_password(test, moulti_test, monkeypatch, request: FixtureRequest):
	"""Functional tests: ensure moulti-askpass returns the right secret and return code."""
	instance_name = f'{getpid()}-askpass-{test["index"]}'
	env = environ | test.get('env', {}) | {'MOULTI_INSTANCE': instance_name}
	moulti_app = Moulti(instance_name=instance_name)
	async with moulti_app.run_test() as pilot:
		proc = await async_exec(*test['cmd'], env=env, stdout=PIPE)
		await pilot.pause(.5)
		await pilot.press(*test.get('input', INPUT_SECRET), 'enter')
		password = await proc.stdout.readline()
		await proc.wait()
		assert proc.returncode == test.get('rc', 0)
		assert password.decode('utf-8').rstrip() == test.get('output', PASSWORD)
		await pilot.press('q')



from textual.app import ComposeResult
from textual.events import Key, Click
from textual.screen import ModalScreen
from textual.widgets import Label

class HelpScreen(ModalScreen):
	"""
	Modal screen presenting the various keybindings available in Moulti.
	"""

	TEXT = """[u]Documentation[/]:
  https://github.com/xavierog/moulti/blob/master/Documentation.md

[u]Keybindings[/]:
  [key]tab[/], [key]shift+tab[/]: next/previous step/component
  [key]ctrl+y[/], [key]ctrl+t[/]: next/previous unanswered question
  [key]enter[/]: collapse/expand the current step
         submit the current input field
  [key]home[/], [key]pgup[/], [key]up[/], [key]down[/], [key]pgdn[/], [key]end[/]: vertical scroll
  [key]ctrl+pgup[/], [key]left[/], [key]right[/], [key]ctrl+pgdn[/]: horizontal scroll
  [key]l[/]: lock scroll: ignore programmatic scrolling
  [key]s[/]: save the current state to the filesystem
  [key]f[/]: maximize the focused log / exit maximized mode
  [key]z[/]: show/hide the console
  [key]o[/], [key]x[/]: collapse/expand existing steps
  [key]O[/], [key]X[/]: collapse/expand new steps
  [key]ctrl+o[/], [key]ctrl+x[/]: collapse/expand both existing and new steps
  [key]d[/]: switch between dark and light modes
  [key]c[/]: copy the contents of the current step
  [key]w[/]: same + preserve colors/include metadata (e.g. the question)
  [key]/[/], [key]?[/]: forward/backward text search
  [key]n[/], [key]N[/]: next/previous text search
  [key]q[/], [key]ctrl+q[/]: quit""".replace('[key]', '[bold #679cff on black]')

	def compose(self) -> ComposeResult:
		label = Label(self.TEXT, id='main_text')
		label.border_title = 'Moulti: help'
		label.border_subtitle = 'Press any key to dismiss this dialog'
		yield label

	def on_key(self, event: Key) -> None:
		"""
		Dismiss this screen if users press any key.
		"""
		event.stop()
		self.dismiss_if_active()

	def on_click(self, event: Click) -> None:
		"""
		Dismiss this screen if users click outside the panel.
		"""
		clicked_widget = self.get_widget_at(event.screen_x, event.screen_y)[0]
		if clicked_widget == self:
			event.stop()
			self.dismiss_if_active()

	def dismiss_if_active(self) -> None:
		if self.is_active:
			self.dismiss()

	DEFAULT_CSS = """
		HelpScreen {
			align: center middle;
			background: $surface 60%;
		}
		#main_text {
			padding: 1;
			background: $surface;
			border: panel $primary;
			border-title-color: #ffffff;
			border-title-align: center;
			border-subtitle-color: $text-muted;
		}
	"""



from argparse import ArgumentParser
from moulti.client import Args, handle_reply, send_to_moulti
from ..collapsiblestep.cli import add_collapsiblestep_options

def add_abstractquestion_options(parser: ArgumentParser, none: bool = False) -> None:
	add_collapsiblestep_options(parser, none)
	parser.add_argument('--text', '-t', type=str, default=None if none else '', help='question text')

def question_get_answer(args: Args) -> None:
	reply = send_to_moulti(args)
	assert reply is not None
	answer = reply.get('answer')
	if answer is not None:
		print(answer)
	handle_reply(reply)



import sys
from argparse import ArgumentParser, BooleanOptionalAction, _SubParsersAction
from moulti.environ import pint
from moulti.client import Args, handle_reply, send_delete
from moulti.client import send_to_moulti_and_handle_reply, send_no_none_to_moulti_and_handle_reply
from moulti.protocol import moulti_connect, send_json_message, recv_json_message
from . import MOULTI_PASS_DEFAULT_READ_SIZE
from ..collapsiblestep.cli import add_collapsiblestep_options

COMMAND = 'step'

def add_cli_arguments(subparsers: _SubParsersAction) -> None:
	# moulti step:
	step_parser = subparsers.add_parser(COMMAND, help='create and manage steps shown by Moulti')
	step_subparsers = step_parser.add_subparsers(required=True)
	add_step_commands(step_subparsers)
	# moulti pass:
	add_pass_command(subparsers)

def add_step_commands(step_subparsers: _SubParsersAction) -> None:
	# moulti step add
	step_add_parser = step_subparsers.add_parser('add', help='add a new step to Moulti')
	step_add_parser.set_defaults(func=send_to_moulti_and_handle_reply, command=COMMAND, action='add')
	step_add_parser.add_argument('id', type=str, help='unique identifier')
	add_step_options(step_add_parser)

	# moulti step update
	step_update_parser = step_subparsers.add_parser('update', help='update an existing Moulti step')
	step_update_parser.set_defaults(func=send_no_none_to_moulti_and_handle_reply, command=COMMAND, action='update')
	step_update_parser.add_argument('id', type=str, help='unique identifier')
	add_step_options(step_update_parser, none=True)

	# moulti step delete
	step_delete_parser = step_subparsers.add_parser('delete', help='delete an existing Moulti step')
	step_delete_parser.set_defaults(func=send_delete, command=COMMAND, action='delete')
	step_delete_parser.add_argument('id', type=str, nargs='+', help='unique identifier')

	# moulti step clear
	step_clear_parser = step_subparsers.add_parser('clear', help='clear the contents of an existing Moulti step')
	step_clear_parser.set_defaults(func=send_to_moulti_and_handle_reply, command=COMMAND, action='clear')
	step_clear_parser.add_argument('id', type=str, help='unique identifier')

	# moulti step append
	step_append_parser = step_subparsers.add_parser('append', help='append the given contents to an existing Moulti step')
	step_append_parser.set_defaults(func=send_to_moulti_and_handle_reply, command=COMMAND, action='append')
	step_append_parser.add_argument('id', type=str, help='unique identifier')
	step_append_parser.add_argument('text', type=str, nargs='+', help='strings to append')

def add_pass_command(subparsers: _SubParsersAction) -> None:
	# moulti pass
	pass_parser = subparsers.add_parser('pass', help='pass standard input to an existing Moulti step')
	pass_parser.set_defaults(func=pass_stdin)
	pass_parser.add_argument('id', type=str, help='unique identifier')
	pass_parser.add_argument('--append', '--no-clear', '-a', dest='append', action='store_true', help='do not clear the target step')
	pass_parser.add_argument('--read-size', '-rs', dest='read_size', type=pint, default=MOULTI_PASS_DEFAULT_READ_SIZE, help='read size')

def add_step_options(parser: ArgumentParser, none: bool = False) -> None:
	"""Options common to step add (with actual default values) and step update (with None default values)."""
	add_collapsiblestep_options(parser, none)
	parser.add_argument('--text', '-t', type=str, default=None if none else '', help='content')
	parser.add_argument('--auto-scroll', action=BooleanOptionalAction, default=None if none else True, help='enable or disable inner auto-scrolling')
	parser.add_argument('--min-height', '-mh', type=pint, default=None if none else  1, help='minimum content height')
	parser.add_argument('--max-height', '-Mh', type=pint, default=None if none else 25, help='maximum content height; 0 to disable')

def pass_stdin(args: Args) -> None:
	"""Pass the stdin file descriptor to Moulti and inject everything read from it into a given step."""
	moulti_socket = moulti_connect()
	step_id = args['id']
	# Clear the target step unless told not to through --append or --no-clear:
	if not args.get('append', False):
		send_json_message(moulti_socket, {'command': COMMAND, 'action': 'clear', 'id': step_id})
		reply, _ = recv_json_message(moulti_socket, 0)
	pass_message = {'command': 'pass', 'id': step_id, 'read_size': args['read_size']}
	send_json_message(moulti_socket, pass_message, [sys.stdin.fileno()])
	# At this stage, the remote Moulti process has the file descriptor and is
	# expected to read lines from it. On our side, we need to remain up &
	# running, otherwise the line-emitting process gets a SIGPIPE.
	# Specifically, we remain up and running until we get a message from the
	# remote Moulti process:
	reply, _ = recv_json_message(moulti_socket, 0)
	moulti_socket.close()
	handle_reply(reply)



from __future__ import annotations
import os
import sys
import subprocess as sp
from datetime import datetime
from pathlib import Path
from typing import Any, TYPE_CHECKING
# pylint: disable=import-error
from ansible.errors import AnsiblePromptInterrupt # type: ignore
from ansible.utils.color import parsecolor # type: ignore
from ansible.utils.display import Display # type: ignore
try: # for Ansible 2.16
	from ansible.utils.display import proxy_display # type: ignore
except ImportError: # for Ansible >= 2.17
	proxy_display = Display._proxy # pylint: disable=protected-access
from ansible.plugins.callback.default import CallbackModule as DefaultCallbackModule # type: ignore
from ansible.utils.unsafe_proxy import wrap_var # type: ignore

if TYPE_CHECKING:
	import collections.abc as c

DOCUMENTATION = '''
    name: moulti
    type: stdout
    short_description: Moulti output for Ansible playbooks
    version_added: N/A
    description:
        - This callback issues `moulti` commands to append plays and tasks to a pre-existing [Moulti](https://github.com/xavierog/moulti) instance.
    extends_documentation_fragment:
      - default_callback
      - result_format_callback
    requirements:
      - set as stdout in configuration
'''

Answer = tuple[str, list[str]]
"""
The first element of the tuple is the answer itself.
The second element is the list of inputquestion widgets created to obtain this answer.
"""

COLLAPSIBLE = ('prompt', 'recap', 'task')
COLLAPSE = [part for part in os.environ.get('MOULTI_ANSIBLE_COLLAPSE', '').split(',') if part in COLLAPSIBLE]
COLLAPSE = ['ansible_' + part for part in COLLAPSE]

def escape_rich_tags(value: str) -> str:
	return value.replace('[', r'\[')

class MoultiDisplay(Display):
	"""
	Variant of Ansible's regular display that:
	- spawns a new Moulti step for some (but not all) banners
	- writes regular messages to the last spawned Moulti step
	"""

	def __init__(self, *args: Any, **kwargs: Any):
		super().__init__(*args, **kwargs)
		self.step_counter = 0
		self.current_step: str|None = None
		self.current_type: str|None = None
		self.start_time = datetime.now()
		self.pipes: dict[str, sp.Popen|None] = {}
		if 'MOULTI_ANSIBLE_NO_TITLE' not in os.environ:
			self.set_title_from_command_line()
		self.is_task: bool = False
		self.task_class: str|None = None
		self.task_class_priority: int = -1
		# Hijack sys.stdout and sys.stderr to harvest as much output as possible:
		sys.stdout = sys.stderr = self

	def __del__(self) -> None:
		# Close the last remaining pipe, if any:
		self.close_current_pipe()

	def moulti(self, args: list[str], **sp_args: Any) -> sp.CompletedProcess:
		return sp.run(['moulti'] + args, shell=False, check=True, **sp_args)

	def set_title_from_command_line(self) -> None:
		simplified_args = []
		for arg in sys.argv:
			if arg.startswith('/'):
				arg = Path(arg).name
			simplified_args.append(arg)
		self.set_title(' '.join(simplified_args))

	def set_title(self, title: str) -> None:
		self.moulti(['set', '--title', title])

	def new_step_id(self) -> str:
		step_id = f'ansible_{os.getpid()}_{self.step_counter}'
		self.step_counter += 1
		return step_id

	def new_widget(self, widget_type: str, title: str, classes: str, *args: str) -> str:
		self.tidy()
		widget_id = self.new_step_id()
		all_classes = 'ansible ' + classes
		final_title = escape_rich_tags(title)
		moulti_args = [widget_type, 'add', widget_id, '--title', final_title, '--classes', all_classes]
		for one_class in classes.split():
			if one_class in COLLAPSE:
				moulti_args += ['--collapsed']
				break
		moulti_args += ['--scroll-on-activity=-1']
		moulti_args += [*args]
		self.moulti(moulti_args)
		self.current_step = widget_id
		self.current_type = widget_type
		return widget_id

	def new_playbook(self, title: str) -> str:
		return self.new_widget('divider', f'PLAYBOOK: {title}', 'ansible_playbook')

	def new_play(self, title: str) -> str:
		return self.new_widget('divider', f'PLAY: {title}', 'ansible_play')

	def date_time(self, date: datetime|None = None) -> str:
		date = datetime.now() if date is None else date
		return date.strftime('%Y-%m-%d %H:%M:%S.%f')[:-3] # milliseconds, not microseconds

	def duration(self, start_date: datetime, end_date: datetime) -> str:
		duration = end_date - start_date
		duration_parts = str(duration).split('.')
		return f'{duration_parts[0]}.{duration_parts[1][:3]}' # milliseconds, not microseconds

	def new_task(self, title: str) -> str:
		task_id = self.new_widget('step', f'{title}', 'ansible_task', '--bottom-text', ' ')
		self.start_time = datetime.now()
		self.moulti(['step', 'update', task_id, '--top-text', f'Started {self.date_time(self.start_time)}'])
		self.is_task = True
		self.task_class = None
		self.task_class_priority = -1
		return task_id

	def new_recap(self, title: str) -> str:
		return self.new_widget('step', f'{title}', 'ansible_recap', '--bottom-text', ' ')

	def tidy(self) -> None:
		date_time = datetime.now()
		self.close_current_pipe()
		if self.current_step and self.is_task:
			duration = self.duration(self.start_time, date_time)
			args = ['--bottom-text', f'Finished {self.date_time(date_time)} - duration: {duration}']
			# Color the previous task green if needed:
			if not self.task_class:
				args += ['--classes', 'ansible ansible_task success']
			self.moulti(['step', 'update', self.current_step, *args])
		self.is_task = False
		self.task_class = None

	def open_pipe(self, step_id: str) -> sp.Popen:
		# pylint: disable=consider-using-with
		pipe = sp.Popen(['moulti', 'pass', step_id], stdin=sp.PIPE, stdout=sp.DEVNULL, stderr=sp.DEVNULL)
		self.pipes[step_id] = pipe
		return pipe

	def get_pipe(self, step_id: str) -> sp.Popen:
		pipe = self.pipes.get(step_id)
		return self.open_pipe(step_id) if pipe is None else pipe

	def close_pipe(self, step_id: str) -> None:
		pipe = self.pipes.get(step_id)
		if pipe is not None:
			pipe.stdin.close() # type: ignore
			self.pipes[step_id] = None
			pipe.wait()

	def close_current_pipe(self) -> None:
		if self.current_step:
			self.close_pipe(self.current_step)

	def do_write(self, step_id: str, msg: str) -> None:
		pipe = self.get_pipe(step_id)
		data = msg
		if not data.endswith('\n'):
			data += '\n'
		pipe.stdin.write(data.encode('utf-8')) # type: ignore
		pipe.stdin.flush() # type: ignore

	@proxy_display
	def display(self, msg: str, color: str | None = None, stderr: bool = False, screen_only: bool = False,
		log_only: bool = False, newline: bool = True, **kwargs: Any
	) -> None:
		_ = stderr
		if log_only or not screen_only:
			super().display(msg=msg, color=color, stderr=False, screen_only=False, log_only=True, newline=newline,
				**kwargs)

		if screen_only or not log_only:
			if not msg:
				return
			task_class, priority = None, -1
			if color is not None:
				color_code = parsecolor(color)
				msg = '\n'.join([f'\033[{color_code}m{t}\033[0m' for t in msg.split('\n')])
				if color in ('red', 'bright red'):
					task_class, priority = 'error', 10
				elif color in ('yellow', 'bright yellow'):
					task_class, priority = 'warning', 5
				elif color in ('green', 'bright green'):
					task_class, priority = 'success', 3
				elif color in ('cyan', 'bright cyan'): # "skipping" and "included" lines
					task_class, priority = 'inactive', 0
			step_id = self.write(msg)
			if task_class is not None and task_class != self.task_class and priority > self.task_class_priority:
				self.task_class = task_class
				self.task_class_priority = priority
				self.moulti(['step', 'update', step_id, '--classes', 'ansible ansible_task ' + self.task_class])

	def banner(self, msg: str, color: str | None = None, _cows: bool = True) -> None:
		if msg.startswith('TASK'):
			self.new_task(f'TASK: {msg[6:-1]}')
		elif msg.startswith('PLAY RECAP'):
			self.new_recap(msg)
		elif msg.startswith('PLAYBOOK'):
			self.new_playbook(msg[10:])
		elif msg.startswith('PLAY'):
			self.new_play(msg[6:-1])
		else:
			super().banner(msg, color, cows=False)

	def write(self, data: str) -> str:
		if self.current_step and self.current_type == 'step':
			step_id = self.current_step
		else:
			step_id = self.new_task('Ansible standard output')
		self.do_write(step_id, data)
		return step_id

	def flush(self) -> None:
		pass

	def close(self) -> None:
		pass

	def get_answer(self, inputquestion_id: str, default: str | None = None) -> str:
		cpe = self.moulti(['inputquestion', 'get-answer', '--wait', inputquestion_id], capture_output=True)
		answer = cpe.stdout.decode('utf-8', errors='surrogate_or_strict').rstrip('\n')
		if not answer and default is not None:
			answer = default
		return answer

	def simple_prompt(self, msg: str, *args: Any, private: bool = False, default: str | None = None) -> Answer:
		cmd_args = list(args)
		if private:
			cmd_args.append('--password')
		if default is not None and not private:
			cmd_args.append(f'--bottom-text=Default value: {default}')
		question_id = self.new_widget('inputquestion', msg, 'ansible_prompt warning', *cmd_args)
		self.moulti(['scroll', question_id, '-1'])
		answer = self.get_answer(question_id, default)
		return (answer, [question_id])

	def confirm_prompt(self, msg: str, *args: Any, private: bool = False, default: str | None = None) -> Answer:
		widgets = []
		confirm = '--top-text=Please enter the value again to confirm:'
		mismatch = '--top-text=[bold]Values entered do not match[/]'
		while True:
			answer1, id1 = self.simple_prompt(msg, *args, private=private, default=default)
			answer2, id2 = self.simple_prompt(msg, *args, confirm, private=private, default=default)
			widgets.extend(id1)
			widgets.extend(id2)
			if answer1 == answer2:
				return (answer1, widgets)
			for widget_id in (id1[0], id2[0]):
				self.moulti(['inputquestion', 'update', '--classes=ansible_prompt error', mismatch, widget_id])

	def prompt(self, msg: str, private: bool = False, default: str | None = None) -> str:
		return self.simple_prompt(msg, private=private, default=default)[0]

	def do_var_prompt(
		self,
		varname: str,
		private: bool = True,
		prompt: str | None = None,
		encrypt: str | None = None,
		confirm: bool = False,
		salt_size: int | None = None,
		salt: str | None = None,
		default: str | None = None,
		unsafe: bool = False,
	) -> str:
		do_prompt = self.confirm_prompt if confirm else self.simple_prompt
		msg = prompt if prompt is not None and prompt != varname else f'Input for {varname}?'
		result, _widgets = do_prompt(msg, private=private, default=default)

		if encrypt:
			# Circular import because encrypt needs a display class
			from ansible.utils.encrypt import do_encrypt # type: ignore # pylint: disable=import-outside-toplevel
			result = do_encrypt(result, encrypt, salt_size=salt_size, salt=salt)
		if unsafe:
			result = wrap_var(result)

		return result

	def byte_to_text(self, byte: bytes) -> str:
		try:
			return byte.decode().upper()
		except UnicodeDecodeError:
			return str(byte)

	def prompt_until(
		self,
		msg: str,
		private: bool = False,
		seconds: int | None = None,
		interrupt_input: c.Container[bytes] | None = None,
		complete_input: c.Container[bytes] | None = None,
	) -> bytes:
		cmd_args = []
		if private:
			cmd_args.append('--password')
		if complete_input is not None and hasattr(complete_input, '__iter__'):
			for key in complete_input:
				key_text = self.byte_to_text(key)
				cmd_args += ['--button', '{input}', 'success', key_text]
		else:
			cmd_args += ['--button', '{input}', 'success', 'Enter']

		if interrupt_input is not None and hasattr(interrupt_input, '__iter__'):
			for key in interrupt_input:
				key_text = self.byte_to_text(key)
				cmd_args += ['--button', ':interrupt:', 'error', key_text]
		else:
			cmd_args += ['--button', ':interrupt:', 'error', 'Interrupt']

		question_id = self.new_widget('question', msg, 'ansible_prompt warning', *cmd_args)
		self.moulti(['scroll', question_id, '-1'])

		command = ['question', 'get-answer', '--wait', question_id]
		try:
			cpe = self.moulti(command, capture_output=True, timeout=seconds)
			answer = cpe.stdout.rstrip(b'\n')
		except sp.TimeoutExpired:
			return b''
		if answer == b':interrupt:':
			raise AnsiblePromptInterrupt('user interrupt')
		return answer

class CallbackModule(DefaultCallbackModule):
	"""
	Same as Ansible's default callback module, with a "MoultiDisplay" instead of a regular Display.
	"""
	CALLBACK_VERSION = 1.0
	CALLBACK_TYPE = 'stdout'
	CALLBACK_NAME = 'moulti'

	def __init__(self) -> None:
		super().__init__()
		# Get a reference to the original Display singleton that was initialized before this file was even loaded:
		original_display = Display()
		# Replace it with our own Display:
		self._display: Display = MoultiDisplay(verbosity=original_display.verbosity)
		# But various other parts of Ansible kept a reference to the original Display singleton.
		# Hijack its do_var_prompt() method (+prompt(), just in case) to handle vars_prompt:
		original_display.do_var_prompt = self._display.do_var_prompt
		original_display.prompt = self._display.prompt
		original_display.prompt_until = self._display.prompt_until



from argparse import _SubParsersAction
from moulti.client import send_delete, send_to_moulti_and_handle_reply, send_no_none_to_moulti_and_handle_reply
from ..abstractquestion.cli import add_abstractquestion_options, question_get_answer

COMMAND = 'buttonquestion' # abridged 'bq' below

def add_cli_arguments(subparsers: _SubParsersAction) -> None:
	# moulti buttonquestion:
	bq_parser = subparsers.add_parser(COMMAND, help='create and manage interactive questions with buttons')
	bq_subparsers = bq_parser.add_subparsers(required=True)
	add_bq_commands(bq_subparsers)

def add_bq_commands(bq_subparsers: _SubParsersAction) -> None:
	# moulti buttonquestion add
	bq_add_parser = bq_subparsers.add_parser('add', help='add a new question to Moulti')
	bq_add_parser.set_defaults(func=send_to_moulti_and_handle_reply, command=COMMAND, action='add')
	bq_add_parser.add_argument('id', type=str, help='unique identifier')
	add_abstractquestion_options(bq_add_parser)
	button_metavar = ('VALUE', 'STYLE', 'LABEL')
	helpmsg = 'add a button; available styles: default, primary, error, warning, success'
	bq_add_parser.add_argument('--button', '-b', required=True, action='append', nargs=3, metavar=button_metavar, help=helpmsg)

	# moulti buttonquestion update
	bq_update_parser = bq_subparsers.add_parser('update', help='update an existing Moulti question')
	bq_update_parser.set_defaults(func=send_no_none_to_moulti_and_handle_reply, command=COMMAND, action='update')
	bq_update_parser.add_argument('id', type=str, help='unique identifier')
	add_abstractquestion_options(bq_update_parser, none=True)

	# moulti buttonquestion get-answer
	helpmsg = 'return the answer provided by the end user'
	bq_getanswer_parser = bq_subparsers.add_parser('get-answer', help=helpmsg)
	bq_getanswer_parser.set_defaults(func=question_get_answer, command=COMMAND, action='get-answer')
	bq_getanswer_parser.add_argument('id', type=str, help='unique identifier')
	bq_getanswer_parser.add_argument('--wait', '-w', action='store_true', help='wait until the answer is available')

	# moulti buttonquestion delete
	bq_delete_parser = bq_subparsers.add_parser('delete', help='delete an existing Moulti question')
	bq_delete_parser.set_defaults(func=send_delete, command=COMMAND, action='delete')
	bq_delete_parser.add_argument('id', type=str, nargs='+', help='unique identifier')



from filecmp import cmpfiles
from pathlib import Path
from os import listdir, getpid
from tempfile import TemporaryDirectory
from moulti.app import Moulti
from .common import moulti_test, wait_script
assert moulti_test

LOAD_DIR = 'tests/data/social-preview'
LOAD_COMMAND = ['moulti', 'load', LOAD_DIR]

def test_load(moulti_test):
	assert moulti_test(command=LOAD_COMMAND)

async def test_save(monkeypatch):
	"""
	Load the same directory as test_load, save it to a temporary directory then
	compare the resulting files against the original ones.
	"""
	with TemporaryDirectory() as working_directory:
		monkeypatch.setenv('MOULTI_SAVE_PATH', working_directory)
		moulti_app = Moulti(command=LOAD_COMMAND, instance_name=f'{getpid()}-test-save')
		async with moulti_app.run_test() as pilot:
			await wait_script(pilot) # wait until everything is loaded
			await pilot.press('s') # save
			await pilot.pause(.5) # enough time to do the job
			await pilot.press('q') # quit
		save_directory = Path(working_directory) / listdir(working_directory)[0]
		# The save directory should contain the exact same files as the load directory:
		save_files = sorted(listdir(save_directory))
		assert save_files == sorted(listdir(LOAD_DIR))
		assert cmpfiles(save_directory, LOAD_DIR, save_files)



from typing import Any, Generator
from rich.console import Console
from rich.style import Style
from rich.text import Text
from textual.strip import Strip
from moulti.helpers import ANSI_ESCAPE_SEQUENCE_STR

class ToLinesMixin:
	def str_to_text(self, string: str) -> Text:
		return Text.from_ansi(string) if ANSI_ESCAPE_SEQUENCE_STR in string else Text(string)

	def text_to_strip(self, text: Text) -> Strip:
		return Strip(text.render(self.app.console)) # type: ignore

	def strip_to_text(self, strip: Strip) -> Text:
		parts: list[str|tuple[str, Style]] = [
			text if style is None else (text, style)
			for text, style, _ in strip
		]
		return Text.assemble(*parts)

	def line_to_plain_text(self, y: int, store: bool = False) -> tuple[str, Text|None]:
		"""
		Return the requested line as plain text.
		If a Text object was instantiated to obtain the returned plain text, this Text object is returned too.
		"""
		line = self.lines[y] # type: ignore
		if isinstance(line, str):
			if ANSI_ESCAPE_SEQUENCE_STR in line:
				line = Text.from_ansi(line)
				if store:
					self.lines[y] = line # type: ignore
				return (line.plain, line)
			return (line, None)
		if isinstance(line, Text):
			return (line.plain, line)
		# At this stage, line is a Strip.
		return (line.text, None)

	def line_to_text(self, y: int, store: bool = False) -> Text:
		line = self.lines[y] # type: ignore
		if isinstance(line, Text):
			return line
		if isinstance(line, str):
			line = self.str_to_text(line)
			if store:
				self.lines[y] = line # type: ignore
			return line
		# At this stage, line is a Strip.
		return self.strip_to_text(line)

	def line_to_strip(self, y: int, store: bool = False) -> Strip:
		line = self.lines[y] # type: ignore
		changed = False
		if isinstance(line, str):
			line = self.str_to_text(line)
		if isinstance(line, Text):
			line = self.text_to_strip(line)
			changed = True
		# At this stage, line is a Strip.
		if changed and store:
			self.lines[y] = line # type: ignore
		return line

	def to_lines(self, keep_styles: bool = True) -> Generator:
		"""
		Our own variant of Strip.render() that does NOT forget to output unstyled segments.
		"""
		console = self.app.console # type: ignore
		color_system = Console()._color_system # pylint: disable=protected-access
		style_render = Style.render
		if keep_styles:
			highlighter = None
			if hasattr(self, 'highlight') and self.highlight and hasattr(self, 'highlighter'):
				highlighter = self.highlighter
			for line in self.lines: # type: ignore
				if isinstance(line, str):
					if highlighter:
						line = highlighter(Text(line))
					else:
						yield line
						continue
				if isinstance(line, Text):
					segments = line.render(console)
				else:
					# At this stage, line is a Strip.
					segments = line
				yield ''.join([
					text
					if style is None
					else style_render(style, text, color_system=color_system)
					for text, style, _ in segments
				])
		else:
			for index in range(len(self.lines)): # type: ignore
				yield self.line_to_plain_text(index)[0]

	def to_file(self, file_descriptor: Any) -> None:
		for line in self.to_lines():
			file_descriptor.write(line)
			file_descriptor.write('\n')
